from typing import TYPE_CHECKING

import pandas as pd
from sqlalchemy.exc import NoResultFound

# TODO Adapt import when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

from ixmp4 import db
from ixmp4.core.exceptions import (
    Forbidden,
    IxmpError,
    NoDefaultRunVersion,
    OperationNotSupported,
)
from ixmp4.core.utils import substitute_type
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.iamc.utils import normalize_df
from ixmp4.db import utils

from .. import base, versions
from ..model import Model, ModelRepository
from ..scenario import Scenario, ScenarioRepository
from .model import Run, RunVersion


class RunVersionRepository(versions.VersionRepository[RunVersion]):
    model_class = RunVersion


class CreateKwargs(TypedDict, total=False):
    scenario_name: str


class RunRepository(
    base.Creator[Run],
    base.Deleter[Run],
    base.Retriever[Run],
    base.Enumerator[Run],
    abstract.RunRepository,
):
    model_class = Run

    models: ModelRepository
    scenarios: ScenarioRepository
    versions: RunVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        self.models = ModelRepository(*args)
        self.scenarios = ScenarioRepository(*args)
        self.versions = RunVersionRepository(*args)

        from .filter import RunFilter

        self.filter_class = RunFilter
        super().__init__(*args)

    def join_auth(self, exc: db.sql.Select[tuple[Run]]) -> db.sql.Select[tuple[Run]]:
        if not utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)
        return exc

    def add(self, model_name: str, scenario_name: str) -> Run:
        # Get or create model
        try:
            exc_model = self.models.select(name=model_name)
            model = self.session.execute(exc_model).scalar_one()
        except NoResultFound:
            model = Model(name=model_name)
            self.session.add(model)

        # Get or create scenario
        try:
            exc_scenario = self.scenarios.select(name=scenario_name)
            scenario = self.session.execute(exc_scenario).scalar_one()
        except NoResultFound:
            scenario = Scenario(name=scenario_name)
            self.session.add(scenario)

        (version,) = (
            # Aggregate MAX over run.version where the run's
            # model and scenario match.
            self.session.query(db.func.max(Run.version))
            .filter(Run.model.has(name=model_name))
            .filter(Run.scenario.has(name=scenario_name))
            .one()
        )
        if version is None:
            version = 0
        version += 1

        run = Run(model=model, scenario=scenario, version=version)
        self.session.add(run)
        return run

    def delete_optimization_data(self, id: int) -> None:
        pass  # TODO: Implement this. @glatterf42

    def delete_meta_data(self, id: int) -> None:
        current_meta = self.backend.meta.tabulate(
            run={"id": id, "default_only": False}
        ).drop(columns=["id", "value", "dtype"])

        if not current_meta.empty:
            self.backend.meta.bulk_delete(current_meta)  # type: ignore[arg-type]

    def delete_iamc_data(self, id: int) -> None:
        current_dps = self.backend.iamc.datapoints.tabulate(
            run={"id": id, "default_only": False}
        )

        if not current_dps.empty:
            self.backend.iamc.datapoints.bulk_delete(current_dps)  # type: ignore[arg-type]

    def delete_checkpoints(self, id: int) -> None:
        checkpoints = self.backend.checkpoints.list(run__id=id)
        for checkpoint in checkpoints:
            self.backend.checkpoints.delete(checkpoint.id)

    @guard("manage")
    def delete(self, id: int) -> None:
        self.lock(id)
        self.delete_meta_data(id)
        self.delete_iamc_data(id)
        self.delete_optimization_data(id)
        self.delete_checkpoints(id)
        return super().delete(id)

    @guard("edit")
    def create(
        self, model_name: str, *args: str, **kwargs: Unpack[CreateKwargs]
    ) -> Run:
        if self.backend.auth_context is not None:
            if not self.backend.auth_context.check_access("edit", model_name):
                raise Forbidden(f"Access to model '{model_name}' denied.")
        run = super().create(model_name, *args, **kwargs)
        return run

    @guard("view")
    def get(self, model_name: str, scenario_name: str, version: int) -> Run:
        exc = self.select(
            model={"name": model_name},
            scenario={"name": scenario_name},
            version=version,
            default_only=False,
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Run.NotFound(
                model=model_name,
                scenario=scenario_name,
                version=version,
            )

    @guard("view")
    def get_by_id(self, id: int, _access_type: str = "view") -> Run:
        exc = self.select(_access_type=_access_type, _skip_filter=True).where(
            self.model_class.id == id
        )

        try:
            obj: Run = self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Run.NotFound(id=id)

        return obj

    @guard("view")
    def get_default_version(self, model_name: str, scenario_name: str) -> Run:
        exc = self.select(
            model={"name": model_name},
            scenario={"name": scenario_name},
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise NoDefaultRunVersion

    @guard("view")
    def tabulate(self, **kwargs: Unpack[abstract.run.EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)

    @guard("view")
    def list(self, **kwargs: Unpack[abstract.run.EnumerateKwargs]) -> list[Run]:
        return super().list(**kwargs)

    @guard("edit")
    def set_as_default_version(self, id: int) -> None:
        run = self.get_by_id(id, _access_type="edit")

        exc = (
            db.update(Run)
            .where(
                Run.model__id == run.model__id,
                Run.scenario__id == run.scenario__id,
                Run.is_default,
            )
            .values(is_default=False)
        )

        with self.backend.event_handler.pause():
            # we dont want to trigger the
            # updated_at fields for this operation.
            self.session.execute(exc)
            self.session.commit()

        run.is_default = True
        self.session.commit()

    @guard("edit")
    def unset_as_default_version(self, id: int) -> None:
        run = self.get_by_id(id, _access_type="edit")

        if not run.is_default:
            raise IxmpError(f"Run with id={id} is not set as the default version.")

        run.is_default = False
        self.session.commit()

    def revert_units(self, transaction__id: int) -> None:
        units_to_revert = self.backend.units.versions.tabulate(
            transaction__id=transaction__id, valid="after_transaction"
        )
        if units_to_revert.empty:
            return

        # NOTE Reverting only affected units requires careful attention to detail to
        # ensure other items' foreignkey constraints are not disturbed. Much easier and
        # safer to revert all; plus we're not running this always and there shouldn't be
        # too many affected units.
        current_units = self.backend.units.tabulate()

        if not current_units.empty:
            self.backend.units.bulk_delete(current_units)

        version_units = self.backend.units.versions.tabulate(
            transaction__id=transaction__id
        )
        if version_units.empty:
            return

        with self.backend.event_handler.pause():
            self.backend.units.bulk_upsert(
                version_units.drop(
                    columns=[
                        "id",
                        "transaction_id",
                        "end_transaction_id",
                        "operation_type",
                    ]
                )
            )

    def _get_or_create_ts(self, run__id: int, df: pd.DataFrame) -> pd.DataFrame:
        df["run__id"] = run__id
        id_cols = ["region", "variable", "unit", "run__id"]
        # create set of unqiue timeseries (if missing)
        ts_df = df[id_cols].drop_duplicates()
        self.backend.iamc.timeseries.bulk_upsert(ts_df, create_related=True)

        # retrieve them again to get database ids
        ts_df = self.backend.iamc.timeseries.tabulate(
            join_parameters=True,
            run={"id": run__id, "default_only": False},
        )
        ts_df = ts_df.rename(columns={"id": "time_series__id"})

        # merge on the identity columns
        return pd.merge(
            df,
            ts_df,
            how="left",
            on=id_cols,
            suffixes=(None, "_y"),
        )

    def revert_iamc_data(self, run: Run, transaction__id: int) -> None:
        current_dps = self.backend.iamc.datapoints.tabulate(
            run={"id": run.id, "default_only": False}
        )

        if not current_dps.empty:
            self.backend.iamc.datapoints.bulk_delete(current_dps)  # type: ignore[arg-type]

        version_dps = self.backend.iamc.datapoints.versions.tabulate(
            transaction__id=transaction__id, run__id=run.id
        )
        if version_dps.empty:
            return

        version_dps = version_dps.drop(
            columns=[
                "id",
                "transaction_id",
                "end_transaction_id",
                "operation_type",
                "time_series__id",
            ]
        )
        version_dps = self._get_or_create_ts(run.id, version_dps).dropna(
            how="all", axis="columns"
        )
        self.backend.iamc.datapoints.bulk_upsert(version_dps)  # type: ignore[arg-type]

    def _restore_deleted_units(self, transaction__id: int) -> None:
        deleted_units = self.backend.units.versions.tabulate(
            transaction__id, valid="after_transaction"
        )
        deleted_units = deleted_units[
            deleted_units["operation_type"] == versions.Operation.DELETE
        ]
        if not deleted_units.empty:
            with self.backend.event_handler.pause():
                self.backend.units.bulk_insert(
                    deleted_units.drop(
                        columns=[
                            "id",
                            "transaction_id",
                            "end_transaction_id",
                            "operation_type",
                        ]
                    )
                )

                # NOTE Need this since there may be no commit() between restoration and
                # using restored units for foreign keys
                self.session.commit()

    def revert_optimization_data(
        self, run: Run, transaction__id: int, revert_platform: bool = False
    ) -> None:
        # TODO Should we figure out a way to say which of the following should actually
        # run? Or just make sure they're noops if nothing's required to do?
        # NOTE Only Scalars store Unit IDs as foreign keys
        self.backend.optimization.scalars.revert(
            transaction__id=transaction__id,
            run__id=run.id,
            revert_platform=revert_platform,
        )
        # NOTE IndexSets need to be reverted before linked items
        self.backend.optimization.indexsets.revert(
            transaction__id=transaction__id, run__id=run.id
        )
        self.backend.optimization.tables.revert(
            transaction__id=transaction__id, run__id=run.id
        )
        self.backend.optimization.parameters.revert(
            transaction__id=transaction__id, run__id=run.id
        )
        self.backend.optimization.equations.revert(
            transaction__id=transaction__id, run__id=run.id
        )
        self.backend.optimization.variables.revert(
            transaction__id=transaction__id, run__id=run.id
        )

    def revert_meta(self, run: Run, transaction__id: int) -> None:
        current_meta = self.backend.meta.tabulate(
            run={"id": run.id, "default_only": False}
        ).drop(columns=["id", "value", "dtype"])

        if not current_meta.empty:
            self.backend.meta.bulk_delete(current_meta)  # type: ignore[arg-type]

        version_meta = self.backend.meta.versions.tabulate(
            transaction__id=transaction__id, run__id=run.id
        )
        if version_meta.empty:
            return

        # remove the transaction_id column
        version_meta = version_meta.drop(
            columns=["transaction_id", "end_transaction_id", "operation_type"]
        )
        version_meta = self.backend.meta.merge_value_columns(version_meta)
        version_meta = version_meta.drop(columns=["dtype", "id"])
        self.backend.meta.bulk_upsert(version_meta)  # type: ignore[arg-type]

    def check_versioning_compatiblity(self) -> None:
        if self.backend.engine.dialect.name != "postgresql":
            raise OperationNotSupported(
                "Versioning is only enabled on 'postgres' platforms..."
            )

    @guard("edit")
    def revert(
        self, id: int, transaction__id: int, revert_platform: bool = False
    ) -> None:
        self.check_versioning_compatiblity()

        run = self.get_by_id(id, _access_type="edit")

        if self.backend.transactions.latest().id == transaction__id:
            # we are already at the right transaction
            return

        if revert_platform:
            self._restore_deleted_units(transaction__id)

        self.revert_iamc_data(run, transaction__id)
        self.revert_optimization_data(
            run, transaction__id, revert_platform=revert_platform
        )
        self.revert_meta(run, transaction__id)

    @guard("edit")
    def lock(self, id: int) -> Run:
        try:
            run = self.get_by_id(id, _access_type="edit")
        except Run.NotFound:
            self.get_by_id(id, _access_type="view")
            raise Forbidden("You may not lock this run.")

        if run.lock_transaction is not None:
            raise Run.IsLocked()
        try:
            latest_transaction_id = self.backend.transactions.latest().id
        except versions.Transaction.NotFound:
            # on sqlite no transactions may exist
            latest_transaction_id = 0
        run.lock_transaction = latest_transaction_id
        self.session.commit()
        return run

    @guard("edit")
    def unlock(self, id: int) -> Run:
        run = self.get_by_id(id, _access_type="edit")
        run.lock_transaction = None
        self.session.commit()
        return run

    # TODO improve performance
    # Suggestion by meksor: write a (single) query to load all objects with run_id to a
    # (single) dataframe, change the run_id, then write all back (does that work at
    # once/with one query?)
    # Alternatively, at least use bulk statements for the optimization items and ensure
    # that commit is only called once at the end (so that either all changes make it or
    # none and to avoid multiple transactions with the DB)
    @guard("edit")
    def clone(
        self,
        run_id: int,
        model_name: str | None = None,
        scenario_name: str | None = None,
        keep_solution: bool = True,
    ) -> Run:
        base_run = self.get_by_id(id=run_id)
        run = self.create(
            model_name=model_name if model_name else base_run.model.name,
            scenario_name=scenario_name if scenario_name else base_run.scenario.name,
        )

        datapoints = normalize_df(
            df=self.backend.iamc.datapoints.tabulate(
                join_parameters=True,
                join_runs=False,
                run={"id": base_run.id, "default_only": False},
            ),
            raw=False,
            join_runs=False,
            join_run_id=False,
        )
        if not datapoints.empty:
            datapoints["run__id"] = run.id
            # TODO This is essentially duplicating core/iamc/data/_get_or_create_ts,
            # which we should probably avoid.
            id_cols = ["region", "variable", "unit", "run__id"]
            # create set of unqiue timeseries (if missing)
            ts_df = datapoints[id_cols].drop_duplicates()
            self.backend.iamc.timeseries.bulk_upsert(ts_df, create_related=True)

            # retrieve them again to get database ids
            ts_df = self.backend.iamc.timeseries.tabulate(
                join_parameters=True,
                run={"id": run.id, "default_only": False},
            )
            ts_df = ts_df.rename(columns={"id": "time_series__id"})

            # merge on the identity columns
            datapoints = pd.merge(
                datapoints, ts_df, how="left", on=id_cols, suffixes=(None, "_y")
            )
            substitute_type(df=datapoints)
            # TODO This function expects a pandera.DataFrame of a certain schema, even
            # though the abstract layer annotates it as pd.DataFrame and it works with
            # one. I'm not simply adjusting the type hint because the function might be
            # called directly, in which case we would like to reject unvalidated
            # pd.DataFrames, I think. However, in this case, the data were validated
            # before and need not be checked again.
            self.backend.iamc.datapoints.bulk_upsert(datapoints)  # type: ignore[arg-type]

        for scalar in base_run.scalars:
            self.backend.optimization.scalars.create(
                run_id=run.id,
                name=scalar.name,
                value=scalar.value,
                unit_name=scalar.unit.name,
            )

        for indexset in base_run.indexsets:
            new_indexset = self.backend.optimization.indexsets.create(
                run_id=run.id, name=indexset.name
            )
            self.backend.optimization.indexsets.add_data(
                id=new_indexset.id, data=indexset.data
            )

        for table in base_run.tables:
            new_table = self.backend.optimization.tables.create(
                run_id=run.id,
                name=table.name,
                constrained_to_indexsets=table.indexset_names,
                column_names=table.column_names,
            )
            self.backend.optimization.tables.add_data(id=new_table.id, data=table.data)

        for parameter in base_run.parameters:
            new_parameter = self.backend.optimization.parameters.create(
                run_id=run.id,
                name=parameter.name,
                constrained_to_indexsets=parameter.indexset_names,
                column_names=parameter.column_names,
            )
            self.backend.optimization.parameters.add_data(
                id=new_parameter.id, data=parameter.data
            )

        for equation in base_run.equations:
            new_equation = self.backend.optimization.equations.create(
                run_id=run.id,
                name=equation.name,
                constrained_to_indexsets=equation.indexset_names,
                column_names=equation.column_names,
            )
            if keep_solution:
                self.backend.optimization.equations.add_data(
                    id=new_equation.id, data=equation.data
                )

        for variable in base_run.variables:
            new_variable = self.backend.optimization.variables.create(
                run_id=run.id,
                name=variable.name,
                constrained_to_indexsets=variable.indexset_names,
                column_names=variable.column_names,
            )
            if keep_solution:
                self.backend.optimization.variables.add_data(
                    id=new_variable.id, data=variable.data
                )

        return run



from typing import Annotated

from fastapi import APIRouter, Body, Depends, Path, Query
from pydantic import Field

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.run.filter import RunFilter
from ixmp4.data.db.run.model import Run

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/runs",
    tags=["runs"],
)


class RunInput(BaseModel):
    name_of_model: Annotated[str, Field(alias="model_name")]
    scenario_name: str


class CloneInput(BaseModel):
    run_id: int
    name_of_model: Annotated[str | None, Field(alias="model_name")] = None
    scenario_name: str | None = Field(None)
    keep_solution: bool = Field(True)


@router.patch("/", response_model=EnumerationOutput[api.Run])
def query(
    filter: RunFilter = Body(RunFilter()),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[Run]:
    return EnumerationOutput(
        results=backend.runs.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.runs.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.Run)
def create(
    run: RunInput,
    backend: Backend = Depends(deps.get_backend),
) -> Run:
    return backend.runs.create(**run.model_dump(by_alias=True))


@router.delete("/{id}/")
def delete(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.runs.delete(id)


@router.post("/{id}/set-as-default-version/")
def set_as_default_version(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.runs.set_as_default_version(id)


@router.post("/{id}/unset-as-default-version/")
def unset_as_default_version(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.runs.unset_as_default_version(id)


@router.get("/{id}/", response_model=api.Run)
def get_by_id(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> Run:
    return backend.runs.get_by_id(id)


class RevertInput(BaseModel):
    transaction__id: int
    revert_platform: bool


@router.post("/{id}/revert/")
def revert(
    id: int = Path(),
    input: RevertInput = Body(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.runs.revert(id, input.transaction__id, input.revert_platform)


@router.post("/{id}/lock/", response_model=api.Run)
def lock(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> Run:
    return backend.runs.lock(id)


@router.post("/{id}/unlock/", response_model=api.Run)
def unlock(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> Run:
    return backend.runs.unlock(id)


@router.post("/clone/", response_model=api.Run)
def clone(
    run: CloneInput,
    backend: Backend = Depends(deps.get_backend),
) -> Run:
    return backend.runs.clone(**run.model_dump(by_alias=True))



import logging
from collections.abc import Generator
from functools import partial, reduce
from typing import TYPE_CHECKING, Any, List, Literal, Sequence, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.equation.model import (
    EquationIndexsetAssociation,
)
from ixmp4.data.db.optimization.equation.repository import EquationRepository
from ixmp4.data.db.optimization.parameter.model import (
    ParameterIndexsetAssociation,
)
from ixmp4.data.db.optimization.parameter.repository import ParameterRepository
from ixmp4.data.db.optimization.table.model import TableIndexsetAssociation
from ixmp4.data.db.optimization.table.repository import TableRepository
from ixmp4.data.db.optimization.variable.model import (
    VariableIndexsetAssociation,
)
from ixmp4.data.db.optimization.variable.repository import VariableRepository
from ixmp4.data.db.utils import map_existing
from ixmp4.db.utils.revert import apply_transaction__id

from .. import base, utils
from .docs import IndexSetDocsRepository
from .model import IndexSet, IndexSetData, IndexSetDataVersion, IndexSetVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

log = logging.getLogger(__name__)


class IndexSetVersionRepository(versions.VersionRepository[IndexSetVersion]):
    model_class = IndexSetVersion


class IndexSetDataVersionRepository(versions.VersionRepository[IndexSetDataVersion]):
    model_class = IndexSetDataVersion

    def select(
        self,
        transaction__id: int | None = None,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        indexset__ids: list[int] | None = None,
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = db.select(
            self.bundle, IndexSetVersion.name.label("indexset")
        ).select_from(self.model_class)

        exc = exc.join(
            IndexSetVersion,
            onclause=IndexSetDataVersion.indexset__id == IndexSetVersion.id,
        )

        _apply_transaction__id = partial(
            apply_transaction__id, transaction__id=transaction__id, valid=valid
        )

        exc = reduce(_apply_transaction__id, {self.model_class, IndexSetVersion}, exc)

        if indexset__ids is not None:
            exc = exc.where(IndexSetDataVersion.indexset__id.in_(indexset__ids))

        exc = db.utils.where_matches_kwargs(exc, model_class=self.model_class, **kwargs)
        return exc.distinct()

    def select_ids(
        self,
        transaction__id: int | None = None,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        indexset__ids: list[int] | None = None,
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = db.select(
            self.model_class.id, IndexSetVersion.name.label("indexset")
        ).select_from(self.model_class)

        exc = exc.join(
            IndexSetVersion,
            onclause=IndexSetDataVersion.indexset__id == IndexSetVersion.id,
        )

        _apply_transaction__id = partial(
            apply_transaction__id, transaction__id=transaction__id, valid=valid
        )

        exc = reduce(_apply_transaction__id, {self.model_class, IndexSetVersion}, exc)

        if indexset__ids is not None:
            exc = exc.where(IndexSetDataVersion.indexset__id.in_(indexset__ids))

        exc = db.utils.where_matches_kwargs(exc, **kwargs)
        return exc.distinct()


class IndexSetDataRepository(base.Reverter[IndexSetData]):
    model_class = IndexSetData
    versions: IndexSetDataVersionRepository

    def __init__(self, backend: "SqlAlchemyBackend") -> None:
        super().__init__(backend)

        self.versions = IndexSetDataVersionRepository(backend)

        from .filter import OptimizationIndexSetDataFilter

        self.filter_class = OptimizationIndexSetDataFilter

    # TODO If we only pass on variables, remove this function implementation
    def revert(
        self, transaction__id: int, correct_versions: pd.DataFrame | None
    ) -> None:
        return super().revert(
            transaction__id=transaction__id, correct_versions=correct_versions
        )


class IndexSetRepository(
    base.Creator[IndexSet],
    base.Deleter[IndexSet],
    base.Retriever[IndexSet],
    base.Enumerator[IndexSet],
    base.Reverter[IndexSet],
    abstract.IndexSetRepository,
):
    model_class = IndexSet
    versions: IndexSetVersionRepository
    _data: IndexSetDataRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.docs = IndexSetDocsRepository(*args)

        from .filter import OptimizationIndexSetFilter

        self.filter_class = OptimizationIndexSetFilter

        self.versions = IndexSetVersionRepository(*args)

        self._data = IndexSetDataRepository(*args)

        self._linked_columns_lookup = {
            "table": (
                TableIndexsetAssociation.table__id,
                TableIndexsetAssociation.indexset__id,
            ),
            "parameter": (
                ParameterIndexsetAssociation.parameter__id,
                ParameterIndexsetAssociation.indexset__id,
            ),
            "equation": (
                EquationIndexsetAssociation.equation__id,
                EquationIndexsetAssociation.indexset__id,
            ),
            "variable": (
                VariableIndexsetAssociation.variable__id,
                VariableIndexsetAssociation.indexset__id,
            ),
        }

    def add(self, run_id: int, name: str) -> IndexSet:
        indexset = IndexSet(run__id=run_id, name=name)
        self.session.add(indexset)
        return indexset

    @guard("view")
    def get(self, run_id: int, name: str) -> IndexSet:
        exc = db.select(IndexSet).where(
            (IndexSet.name == name) & (IndexSet.run__id == run_id)
        )
        try:
            return self.session.execute(exc).scalar_one()
        except db.NoResultFound:
            raise IndexSet.NotFound

    @guard("view")
    def get_by_id(self, id: int) -> IndexSet:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise IndexSet.NotFound(id=id)

        return obj

    @guard("edit")
    def create(self, run_id: int, name: str) -> IndexSet:
        return super().create(run_id=run_id, name=name)

    @guard("edit")
    def delete(self, id: int) -> None:
        super().delete(id=id)

    @guard("view")
    def list(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> list[IndexSet]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack["base.EnumerateKwargs"]) -> pd.DataFrame:
        return super().tabulate(**kwargs).rename(columns={"_data_type": "data_type"})

    @guard("edit")
    def add_data(
        self, id: int, data: float | int | str | List[float] | List[int] | List[str]
    ) -> None:
        _data = data if isinstance(data, list) else [data]

        if len(_data) == 0:
            return  # nothing to be done

        indexset = self.get_by_id(id=id)

        bulk_insert_enabled_data = [{"value": str(d)} for d in _data]
        try:
            self.session.execute(
                db.insert(IndexSetData).values(indexset__id=id),
                bulk_insert_enabled_data,
            )
        except db.IntegrityError as e:
            self.session.rollback()
            raise indexset.DataInvalid from e

        # Due to _data's limitation above, __name__ will always be that
        indexset._data_type = cast(
            Literal["float", "int", "str"], type(_data[0]).__name__
        )

        # Flush & expire session to refresh IndexSets stored in it
        self.session.commit()

    @guard("edit")
    def remove_data(
        self,
        id: int,
        data: float | int | str | List[float] | List[int] | List[str],
        remove_dependent_data: bool = True,
    ) -> None:
        if not bool(data):
            return

        _data = [str(d) for d in data] if isinstance(data, list) else [str(data)]

        if remove_dependent_data:
            repos: dict[
                Literal["table", "parameter", "equation", "variable"],
                TableRepository
                | ParameterRepository
                | EquationRepository
                | VariableRepository,
            ] = {
                "table": self.backend.optimization.tables,
                "parameter": self.backend.optimization.parameters,
                "equation": self.backend.optimization.equations,
                "variable": self.backend.optimization.variables,
            }
            for kind, ids in self.find_all_linked_item_ids(id=id):
                self.remove_invalid_linked_data(
                    repo=repos[kind],
                    ids=ids,
                    indexset_name=self.get_by_id(id=id).name,
                    data=_data,
                )

        result = self.session.execute(
            db.delete(IndexSetData).where(
                IndexSetData.indexset__id == id,
                IndexSetData.value.in_(_data),
            )
        )

        if result.rowcount == 0:
            log.info(f"No data were removed! Are {data} registered to IndexSet {id}?")
            return None
        elif result.rowcount != len(_data):
            log.info(f"Not all items in `data` were registered for IndexSet {id}!")

        # NOTE We are currently not resetting indexset._data_type even if all data are
        # removed from that indexset.

        # Flush & expire session to refresh IndexSets stored in it
        self.session.commit()

    def _find_linked_item_ids(
        self, id: int, item_kind: Literal["table", "parameter", "equation", "variable"]
    ) -> Sequence[int]:
        """Finds all items of `item_kind` linked to an IndexSet in `session.

        Parameters
        ----------
        id : int
            The id of the IndexSet we are looking for.
        item_kind : Literal["table", "parameter", "equation", "variable"]
            The type of item we are looking for.

        Returns
        -------
        list of int
            A list of ids of items of `item_kind` linked to the IndexSet.
        """

        column_clause, compare_column = self._linked_columns_lookup[item_kind]

        statement = db.select(column_clause).where(compare_column == id)

        return self.session.scalars(statement).all()

    @guard("view")
    def find_all_linked_item_ids(
        self, id: int
    ) -> Generator[
        tuple[Literal["table", "parameter", "equation", "variable"], Sequence[int]],
        Any,
        None,
    ]:
        """Finds all optimization items linked to an IndexSet.

        This is done by iterating over all possible kinds and yielding the ids of linked
        items.

        Parameters
        ----------
        id : int
            The id of the IndexSet we are looking for.

        Yields
        ------
        (item kind, list of ids)
            A tuple with the item kind being one of
            {'table', 'parameter', 'equation', 'variable'} and a list of integer ids
            representing linked items.
        """
        item_kinds: set[Literal["table", "parameter", "equation", "variable"]] = {
            "table",
            "parameter",
            "equation",
            "variable",
        }
        for kind in item_kinds:
            yield (kind, self._find_linked_item_ids(id=id, item_kind=kind))

    @guard("edit")
    def remove_invalid_linked_data(
        self,
        repo: TableRepository
        | ParameterRepository
        | EquationRepository
        | VariableRepository,
        ids: Sequence[int],
        indexset_name: str,
        data: List[str],
    ) -> None:
        """Remove invalid data from linked optimization items.

        Parameters
        ----------
        repo : TableRepository | ParameterRepository | EquationRepository | VariableRepository
            The repository including the linked items.
        ids : Sequence[int]
            The IDs of items linked to the IndexSet with `indexset_name`.
        indexset_name : str
            The name of the IndexSet from which data is to be removed.
        data : list[str]
            The data to be removed from `indexset_name` in str format.
        """  # noqa: E501
        for item in repo.list(id__in=ids):
            # Convert existing data for manipulation
            df = pd.DataFrame(item.data)

            if df.empty:
                continue  # nothing to do

            # Identify column/dimension names to target; item and name must be linked
            columns = utils._find_columns_linked_to_indexset(
                item=item, name=indexset_name
            )

            # Prepare data template to exclude
            invalid_data = {columns[i]: data for i in range(len(columns))}

            # Prepare stored data to remove
            # NOTE if any linked column is not of type str, this may be incorrect
            remove_data = df[df[columns].isin(invalid_data).any(axis=1)]

            repo.remove_data(item.id, remove_data)

    def _map_indexsets_on_revert(self, df: pd.DataFrame) -> pd.DataFrame:
        """Maps foreign-key-indexset-ids to the ids of restored IndexSets."""

        df, missing = map_existing(
            df,
            existing_df=self.tabulate(name__in=df["indexset"]),
            join_on=("name", "indexset"),
            map=("id", "indexset__id"),
            suffixes=("_old", None),
        )
        if len(missing) > 0:
            raise IndexSet.NotFound(", ".join(missing))

        return df

    @guard("edit")
    def revert(self, transaction__id: int, run__id: int) -> None:
        super().revert(transaction__id=transaction__id, run__id=run__id)

        # Revert IndexSetData: map indexset ids (and their data) from before
        # transaction__id to after
        indexset_map_subquery = utils.create_id_map_subquery(
            repo=self, transaction__id=transaction__id, run__id=run__id
        )

        columns = db.utils.collect_columns_to_select(
            db.utils.get_columns(self._data.versions.model_class),
            exclude={"indexset__id"},
        )

        select_correct_versions = (
            db.select(
                indexset_map_subquery.c.new_id.label("indexset__id"), *columns.values()
            )
            .select_from(self._data.versions.model_class)
            .join(
                indexset_map_subquery,
                self._data.versions.model_class.indexset__id
                == indexset_map_subquery.c.old_id,
            )
        )

        select_correct_versions = apply_transaction__id(
            exc=select_correct_versions,
            model_class=self._data.versions.model_class,
            transaction__id=transaction__id,
        ).order_by(self._data.versions.model_class.id.asc())

        correct_versions = self.tabulate_query(select_correct_versions)

        # Revert with corrected (restored) indexset_ids
        self._data.revert(
            transaction__id=transaction__id, correct_versions=correct_versions
        )



from typing import TYPE_CHECKING, Any, cast

import pandas as pd
import pandera as pa
from pandera.pandas import DataFrameModel
from pandera.typing import DataFrame, Series
from sqlalchemy.exc import NoResultFound

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict, Unpack

from ixmp4 import db
from ixmp4.core.decorators import check_types
from ixmp4.core.exceptions import InvalidRunMeta
from ixmp4.data import abstract
from ixmp4.data.abstract.meta import EnumerateKwargs as AbstractEnumerateKwargs
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db.model import Model
from ixmp4.data.db.run import Run
from ixmp4.data.db.scenario import Scenario
from ixmp4.db.filters import BaseFilter

from .. import base, versions
from .model import RunMetaEntry, RunMetaEntryVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

ILLEGAL_META_KEYS = {"model", "scenario", "id", "version", "is_default"}


class RemoveRunMetaEntryFrameSchema(DataFrameModel):
    key: Series[pa.String] = pa.Field(coerce=True)
    run__id: Series[pa.Int] = pa.Field(coerce=True)

    class Config:
        strict = True
        coerce = True


class AddRunMetaEntryFrameSchema(RemoveRunMetaEntryFrameSchema):
    value: Series[pa.Object] = pa.Field(coerce=True)


class UpdateRunMetaEntryFrameSchema(AddRunMetaEntryFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


class EnumerateKwargs(AbstractEnumerateKwargs, total=False):
    _filter: BaseFilter


class CreateKwargs(TypedDict, total=False):
    run__id: int
    key: str
    value: abstract.annotations.PrimitiveTypes


class RunMetaEntryVersionRepository(versions.VersionRepository[RunMetaEntryVersion]):
    model_class = RunMetaEntryVersion


class RunMetaEntryRepository(
    base.Creator[RunMetaEntry],
    base.Enumerator[RunMetaEntry],
    base.BulkUpserter[RunMetaEntry],
    base.BulkDeleter[RunMetaEntry],
    abstract.RunMetaEntryRepository,
):
    model_class = RunMetaEntry
    versions: RunMetaEntryVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)
        self.versions = RunMetaEntryVersionRepository(*args)

        from .filter import RunMetaEntryFilter

        self.filter_class = RunMetaEntryFilter

    def add(
        self, run__id: int, key: str, value: abstract.annotations.PrimitiveTypes
    ) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        if key in ILLEGAL_META_KEYS:
            raise InvalidRunMeta("Illegal meta key: " + key)
        entry = RunMetaEntry(run__id=run__id, key=key, value=value)
        self.session.add(entry)
        return entry

    def check_df_access(self, df: pd.DataFrame) -> None:
        if self.backend.auth_context is not None:
            ts_ids = cast(set[int], set(df["run__id"].unique().tolist()))
            self.backend.runs.check_access(
                ts_ids,
                access_type="edit",
                is_default=None,
                default_only=False,
            )

    @guard("edit")
    def create(
        self, *args: abstract.annotations.PrimitiveTypes, **kwargs: Unpack[CreateKwargs]
    ) -> RunMetaEntry:
        return super().create(*args, **kwargs)

    @guard("view")
    def get(self, run__id: int, key: str) -> RunMetaEntry:
        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {run__id},
                access_type="view",
                is_default=None,
                default_only=False,
            )

        exc = self.select(run_id=run__id, key=key)

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise RunMetaEntry.NotFound(run__id=run__id, key=key)

    @guard("edit")
    def delete(self, id: int) -> None:
        exc = db.select(RunMetaEntry).where(RunMetaEntry.id == id)

        try:
            meta = self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise RunMetaEntry.NotFound(id=id)

        if self.backend.auth_context is not None:
            self.backend.runs.check_access(
                {meta.run__id},
                access_type="edit",
                is_default=None,
                default_only=False,
            )

        self.session.delete(meta)
        self.session.commit()

    def join_auth(
        self, exc: db.sql.Select[tuple[RunMetaEntry]]
    ) -> db.sql.Select[tuple[RunMetaEntry]]:
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, RunMetaEntry.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return super().join_auth(exc)

    def select_with_run_index(self) -> db.sql.Select[tuple[str, str, int, Any]]:
        _exc = db.select(
            Model.name.label("model_name"),
            Scenario.name.label("scenario_name"),
            Run.version.label("version"),
            self.bundle,
        ).select_from(self.model_class)

        return (
            _exc.join(Run, Run.id == RunMetaEntry.run__id)
            .join(Model, onclause=Model.id == Run.model__id)
            .join(Scenario, onclause=Scenario.id == Run.scenario__id)
        )

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[RunMetaEntry]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(
        self,
        join_run_index: bool = False,
        _raw: bool | None = False,
        **kwargs: Unpack[EnumerateKwargs],
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(**kwargs)

        if join_run_index:
            _exc = self.select_with_run_index()
            df = super().tabulate(_exc=_exc, **kwargs)
            df.drop(columns="run__id", inplace=True)
            df.rename(
                columns={"model_name": "model", "scenario_name": "scenario"},
                inplace=True,
            )
            index_columns = ["model", "scenario", "version"]
        else:
            df = super().tabulate(**kwargs)
            index_columns = ["run__id"]

        if df.empty:
            return pd.DataFrame(
                [], columns=index_columns + ["id", "dtype", "key", "value"]
            )

        return self.merge_value_columns(df)

    def merge_value_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        def map_value_column(df: pd.DataFrame) -> pd.DataFrame:
            type_str = df.name
            type_ = RunMetaEntry.Type(type_str)
            col = RunMetaEntry._column_map[type_]
            df["value"] = df[col]
            df["dtype"] = type_str
            return df.drop(columns=RunMetaEntry._column_map.values())

        # ensure compatibility with pandas y 2.2
        # TODO remove legacy-handling when dropping support for pandas < 2.2
        apply_args = (
            dict()
            if pd.__version__[0:3] in ["2.0", "2.1"]
            else dict(include_groups=False)
        )

        return df.groupby("dtype", group_keys=False).apply(
            map_value_column, **apply_args
        )

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddRunMetaEntryFrameSchema]) -> None:
        if illegal_keys := (set(df.key.values) & ILLEGAL_META_KEYS):
            raise InvalidRunMeta("Illegal meta key(s): " + ", ".join(illegal_keys))

        self.check_df_access(df)
        df["dtype"] = df["value"].map(type).map(RunMetaEntry.Type.from_pytype)

        for type_, type_df in df.groupby("dtype"):
            # This cast should always be a no-op
            col = RunMetaEntry._column_map[cast(str, type_)]
            null_cols = set(RunMetaEntry._column_map.values()) - set([col])
            type_df["dtype"] = type_df["dtype"].map(lambda x: x.value)
            type_df = type_df.rename(columns={"value": col})

            # ensure all other columns are overwritten
            for nc in null_cols:
                type_df[nc] = None

            super().bulk_upsert(type_df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveRunMetaEntryFrameSchema]) -> None:
        self.check_df_access(df)
        super().bulk_delete(df)



from typing import Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract import annotations

from . import base


class EnumerateKwargs(
    annotations.HasRunIdFilter,
    annotations.HasTransactionIdFilter,
    total=False,
):
    pass


class Checkpoint(base.BaseModel, Protocol):
    """Run checkpoint model.
    Represents a run's state (the data it holds) at a particular transaction.
    Used to roll back to a previous state of the run."""

    run__id: int
    "Id of associated run."

    transaction__id: int | None
    "Id of associated transaction. `None` on sqlite databases."

    message: str
    "Checkpoint message."

    def __str__(self) -> str:
        return f"<Checkpoint {self.id} message={self.message}>"


class CheckpointRepository(
    base.Creator,
    base.Deleter,
    base.Enumerator,
    Protocol,
):
    def create(self, run__id: int, message: str) -> Checkpoint:
        """Creates a checkpoint.

        Parameters
        ----------
        run__id: int
            Id of associated run.

        transaction__id: int
            Id of associated transaction.

        message: str
            Checkpoint message.

        Raises
        ------
        :class:`ixmp4.data.abstract.checkpoint.Checkpoint.NotUnique`:
            If the checkpoints combination of `run__id` and
            `transaction__id` is not unique.

        Returns
        -------
        :class:`ixmp4.data.abstract.Model`:
            The created model.
        """
        ...

    def delete(self, id: int) -> Checkpoint:
        """Deletes a checkpoint by its id.

        Parameters
        ----------
        id : int
            The id of the checkpoint.

        Raises
        ------
        :class:`ixmp4.data.abstract.Checkpoint.NotFound`:
            If the checkpoint does not exist.

        """
        ...

    def get_by_id(self, id: int) -> Checkpoint:
        """Retrieves a checkpoint by its id.

        Parameters
        ----------
        id : int
            The id of the checkpoint.

        Raises
        ------
        :class:`ixmp4.data.abstract.Checkpoint.NotFound`:
            If the checkpoint does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.Checkpoint`:
            The retrieved checkpoint.
        """
        ...

    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[Checkpoint]:
        r"""Lists checkpoints by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.checkpoint.filter.ModelFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.Model`]:
            List of Model.
        """
        ...

    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        r"""Tabulate checkpoints by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.checkpoint.filter.ModelFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name

        """
        ...



from typing import Any, ClassVar

from ixmp4.db import Session, filters, sql, typing_column, utils

from .. import Run
from ..base import SelectType
from ..iamc import TimeSeries
from .model import ModelFilter
from .scenario import ScenarioFilter


class RunFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    version: filters.Integer | None = filters.Field(None)
    default_only: filters.Boolean = filters.Field(True)
    is_default: filters.Boolean | None = filters.Field(None)
    model: ModelFilter | None = filters.Field(None)
    scenario: ScenarioFilter | None = filters.Field(None)

    sqla_model: ClassVar[type] = Run

    def filter_default_only(
        self,
        exc: sql.Select[tuple[Run]],
        c: typing_column[Any],  # Any since it is unused
        v: bool,
        session: Session | None = None,
    ) -> sql.Select[tuple[Run]]:
        return exc.where(Run.is_default) if v else exc

    def join(self, exc: SelectType, session: Session | None = None) -> SelectType:
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        return exc



# type: ignore
"""add audit info

Revision ID: c29289ced488
Revises: 081bbda6bb7b
Create Date: 2024-07-31 11:13:39.889253

"""

import sqlalchemy as sa
from alembic import op

from ixmp4.db.migrations import utils

# Revision identifiers, used by Alembic.
revision = "c29289ced488"
down_revision = "081bbda6bb7b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.add_column(sa.Column("created_at", sa.DateTime(), nullable=True))
        batch_op.add_column(
            sa.Column("created_by", sa.String(length=255), nullable=True)
        )

    with utils.pause_foreign_key_checks():
        with op.batch_alter_table("region", schema=None) as batch_op:
            batch_op.alter_column(
                "name",
                existing_type=sa.VARCHAR(length=1023),
                type_=sa.String(length=255),
                existing_nullable=False,
            )

    with op.batch_alter_table("run", schema=None) as batch_op:
        batch_op.add_column(sa.Column("updated_at", sa.DateTime(), nullable=True))
        batch_op.add_column(
            sa.Column("updated_by", sa.String(length=255), nullable=True)
        )
        batch_op.add_column(sa.Column("created_at", sa.DateTime(), nullable=True))
        batch_op.add_column(
            sa.Column("created_by", sa.String(length=255), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run", schema=None) as batch_op:
        batch_op.drop_column("created_by")
        batch_op.drop_column("created_at")
        batch_op.drop_column("updated_by")
        batch_op.drop_column("updated_at")

    with utils.pause_foreign_key_checks():
        with op.batch_alter_table("region", schema=None) as batch_op:
            batch_op.alter_column(
                "name",
                existing_type=sa.String(length=255),
                type_=sa.VARCHAR(length=1023),
                existing_nullable=False,
            )

    with op.batch_alter_table("optimization_column", schema=None) as batch_op:
        batch_op.drop_column("created_by")
        batch_op.drop_column("created_at")

    # ### end Alembic commands ###



from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, cast

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data.abstract import Docs as DocsModel
from ixmp4.data.abstract import Equation as EquationModel

from .base import (
    Creator,
    Deleter,
    Lister,
    OptimizationBaseModelFacade,
    Retriever,
    Tabulator,
)

if TYPE_CHECKING:
    from ixmp4.core.run import Run

    from . import InitKwargs


class Equation(OptimizationBaseModelFacade):
    _model: EquationModel
    NotFound: ClassVar = EquationModel.NotFound
    NotUnique: ClassVar = EquationModel.NotUnique

    @property
    def id(self) -> int:
        return self._model.id

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def run_id(self) -> int:
        return self._model.run__id

    @property
    def data(self) -> dict[str, list[float] | list[int] | list[str]]:
        return self._model.data

    def add(self, data: dict[str, Any] | pd.DataFrame) -> None:
        """Adds data to the Equation."""
        self._run.require_lock()
        self.backend.optimization.equations.add_data(id=self._model.id, data=data)
        self._model = self.backend.optimization.equations.get(
            run_id=self._model.run__id, name=self._model.name
        )

    # TODO Make name of these functions consistent across items
    def remove_data(self, data: dict[str, Any] | pd.DataFrame | None = None) -> None:
        """Removes data from the Equation.

        If `data` is `None` (the default), remove all data. Otherwise, data must specify
        all indexed columns. All other keys/columns are ignored.
        """
        self._run.require_lock()
        self.backend.optimization.equations.remove_data(id=self._model.id, data=data)
        self._model = self.backend.optimization.equations.get(
            run_id=self._model.run__id, name=self._model.name
        )

    @property
    def levels(self) -> list[float]:
        return cast(list[float], self._model.data.get("levels", []))

    @property
    def marginals(self) -> list[float]:
        return cast(list[float], self._model.data.get("marginals", []))

    @property
    def indexset_names(self) -> list[str] | None:
        return self._model.indexset_names

    @property
    def column_names(self) -> list[str] | None:
        return self._model.column_names

    @property
    def created_at(self) -> datetime | None:
        return self._model.created_at

    @property
    def created_by(self) -> str | None:
        return self._model.created_by

    @property
    def docs(self) -> str | None:
        try:
            return self.backend.optimization.equations.docs.get(self.id).description
        except DocsModel.NotFound:
            return None

    @docs.setter
    def docs(self, description: str | None) -> None:
        if description is None:
            self.backend.optimization.equations.docs.delete(self.id)
        else:
            self.backend.optimization.equations.docs.set(self.id, description)

    @docs.deleter
    def docs(self) -> None:
        try:
            self.backend.optimization.equations.docs.delete(self.id)
        # TODO: silently failing
        except DocsModel.NotFound:
            return None

    def __str__(self) -> str:
        return f"<Equation {self.id} name={self.name}>"


class EquationRepository(
    Creator[Equation, EquationModel],
    Deleter[Equation, EquationModel],
    Retriever[Equation, EquationModel],
    Lister[Equation, EquationModel],
    Tabulator[Equation, EquationModel],
):
    def __init__(self, _run: "Run", **kwargs: Unpack["InitKwargs"]) -> None:
        super().__init__(_run=_run, **kwargs)
        self._backend_repository = self.backend.optimization.equations
        self._model_type = Equation

    def create(
        self,
        name: str,
        constrained_to_indexsets: list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Equation:
        return super().create(
            name=name,
            constrained_to_indexsets=constrained_to_indexsets,
            column_names=column_names,
        )



# type: ignore
"""add checkpoints and run.lock_transaction column

Revision ID: bc2e195788ba
Revises: 0b45de1602a6
Create Date: 2025-03-13 16:57:24.338393

"""

import sqlalchemy as sa
from alembic import op

from ixmp4.db.migrations import utils

# Revision identifiers, used by Alembic.
revision = "bc2e195788ba"
down_revision = "0b45de1602a6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "checkpoint",
        sa.Column("run__id", sa.Integer(), nullable=False),
        sa.Column("transaction__id", sa.Integer(), nullable=False),
        sa.Column("message", sa.String(), nullable=False),
        sa.Column(
            "id",
            sa.Integer(),
            sa.Identity(always=False, on_null=True, start=1, increment=1),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["run__id"], ["run.id"], name=op.f("fk_checkpoint_run__id_run")
        ),
        sa.ForeignKeyConstraint(
            ["transaction__id"],
            ["transaction.id"],
            name=op.f("fk_checkpoint_transaction__id_transaction"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_checkpoint")),
    )
    with op.batch_alter_table("checkpoint", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_checkpoint_run__id"), ["run__id"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_checkpoint_transaction__id"),
            ["transaction__id"],
            unique=False,
        )

    with op.batch_alter_table("run", schema=None) as batch_op:
        batch_op.add_column(sa.Column("lock_transaction", sa.Integer(), nullable=True))
        batch_op.create_index(
            batch_op.f("ix_run_lock_transaction"), ["lock_transaction"], unique=False
        )

    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "lock_transaction", sa.Integer(), autoincrement=False, nullable=True
            )
        )
        batch_op.create_index(
            batch_op.f("ix_run_version_lock_transaction"),
            ["lock_transaction"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("run_version", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_run_version_lock_transaction"))
        batch_op.drop_column("lock_transaction")

    with utils.pause_foreign_key_checks():
        with op.batch_alter_table("run", schema=None) as batch_op:
            batch_op.drop_index(batch_op.f("ix_run_lock_transaction"))
            batch_op.drop_column("lock_transaction")

    with op.batch_alter_table("checkpoint", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_checkpoint_transaction__id"))
        batch_op.drop_index(batch_op.f("ix_checkpoint_run__id"))

    op.drop_table("checkpoint")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Any, ClassVar

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationDataValidationError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db import versions
from ixmp4.data.db.optimization.associations import (
    BaseIndexSetAssociation,
    BaseIndexSetAssociationVersion,
)

from .. import base, utils

if TYPE_CHECKING:
    from .. import IndexSet


class ParameterIndexsetAssociation(BaseIndexSetAssociation):
    __tablename__ = "opt_par_idx_association"

    parameter__id: types.ParameterId
    parameter: types.Mapped["Parameter"] = db.relationship(
        back_populates="_parameter_indexset_associations"
    )


class Parameter(base.BaseModel):
    __tablename__ = "opt_par"

    # NOTE: These might be mixin-able, but would require some abstraction
    NotFound: ClassVar = abstract.Parameter.NotFound
    NotUnique: ClassVar = abstract.Parameter.NotUnique
    DataInvalid: ClassVar = OptimizationDataValidationError
    DeletionPrevented: ClassVar = abstract.Parameter.DeletionPrevented

    run__id: types.RunId
    data: types.JsonDict = db.Column(db.JsonType, nullable=False, default={})

    @db.validates("data")
    def validate_data(self, key: Any, data: dict[str, Any]) -> dict[str, Any]:
        if not bool(data):
            return data
        # This case can only arise when all keys have been deleted
        elif not bool(set(data.keys()) - {"values", "units"}):
            return {}
        utils.validate_data(
            host=self,
            data=data,
            indexsets=self._indexsets,
            column_names=self.column_names,
        )
        return data

    _parameter_indexset_associations: types.Mapped[
        list[ParameterIndexsetAssociation]
    ] = db.relationship(
        back_populates="parameter",
        cascade="all, delete-orphan",
        order_by="ParameterIndexsetAssociation.id",
        passive_deletes=True,
    )

    _indexsets: db.AssociationProxy[list["IndexSet"]] = db.association_proxy(
        "_parameter_indexset_associations", "indexset"
    )
    _column_names: db.AssociationProxy[list[str | None]] = db.association_proxy(
        "_parameter_indexset_associations", "column_name"
    )

    @property
    def indexset_names(self) -> list[str]:
        return [indexset.name for indexset in self._indexsets]

    @property
    def column_names(self) -> list[str] | None:
        names = [name for name in self._column_names if name]
        return names if bool(names) else None

    updateable_columns = ["data"]

    __table_args__ = (db.UniqueConstraint("name", "run__id"),)


class ParameterVersion(versions.DefaultVersionModel):
    __tablename__ = "opt_par_version"

    name: types.String = db.Column(db.String(255), nullable=False)
    run__id: db.MappedColumn[int] = db.Column(db.Integer, nullable=False, index=True)

    data: types.JsonDict = db.Column(db.JsonType, nullable=False, default={})

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


class ParameterIndexsetAssociationVersion(BaseIndexSetAssociationVersion):
    __tablename__ = "opt_par_idx_association_version"

    parameter__id: db.MappedColumn[int] = db.Column(
        db.Integer, nullable=False, index=True
    )


version_triggers = versions.PostgresVersionTriggers(
    Parameter.__table__, ParameterVersion.__table__
)
data_version_triggers = versions.PostgresVersionTriggers(
    ParameterIndexsetAssociation.__table__,
    ParameterIndexsetAssociationVersion.__table__,
)



from typing import TYPE_CHECKING, ClassVar

import pandas as pd
from sqlalchemy.exc import NoResultFound

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.data import abstract, types
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import mixins

from ..unit import Unit

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend

from .. import versions
from . import base
from .variable import Variable


class Measurand(base.BaseModel, mixins.HasCreationInfo):
    __tablename__ = "iamc_measurand"
    NotFound: ClassVar = abstract.Measurand.NotFound
    NotUnique: ClassVar = abstract.Measurand.NotUnique
    DeletionPrevented: ClassVar = abstract.Measurand.DeletionPrevented

    __table_args__ = (db.UniqueConstraint("variable__id", "unit__id"),)

    variable__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("iamc_variable.id"), nullable=False, index=True
    )
    variable: types.Mapped[Variable] = db.relationship(
        "Variable",
        backref="measurands",
        foreign_keys=[variable__id],
        lazy="select",
    )

    unit__id: types.Integer = db.Column(
        db.Integer, db.ForeignKey("unit.id"), nullable=False, index=True
    )
    unit: types.Mapped[Unit] = db.relationship(
        "Unit",
        backref="measurands",
        foreign_keys=[unit__id],
        lazy="select",
    )


class MeasurandVersion(versions.DefaultVersionModel):
    __tablename__ = "iamc_measurand_version"
    variable__id: types.Integer = db.Column(nullable=False, index=True)
    unit__id: types.Integer = db.Column(nullable=False, index=True)

    created_at: types.DateTime = db.Column(nullable=True)
    created_by: types.Username


version_triggers = versions.PostgresVersionTriggers(
    Measurand.__table__, MeasurandVersion.__table__
)


class MeasurandRepository(
    base.Creator[Measurand],
    base.Retriever[Measurand],
    base.Enumerator[Measurand],
    abstract.MeasurandRepository,
):
    model_class = Measurand

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        super().__init__(*args)

    @guard("view")
    def get(self, variable_name: str, unit__id: int) -> Measurand:
        exc = (
            db.select(Measurand)
            .join(Measurand.variable)
            .where(Measurand.unit__id == unit__id)
            .where(Variable.name == variable_name)
        )

        try:
            return self.session.execute(exc).scalar_one()
        except NoResultFound:
            raise Measurand.NotFound

    def add(self, variable_name: str, unit__id: int) -> Measurand:
        q_exc = db.select(Variable).where(Variable.name == variable_name)
        try:
            variable = self.session.execute(q_exc).scalar_one()
        except NoResultFound:
            variable = Variable(name=variable_name)

        measurand = Measurand(variable=variable, unit__id=unit__id)
        self.session.add(measurand)
        self.session.commit()
        return measurand

    @guard("edit")
    def create(self, *args: Unpack[tuple[str, int]]) -> Measurand:
        return super().create(*args)

    def select(self) -> db.sql.Select[tuple[Measurand]]:
        return db.select(Measurand)

    @guard("view")
    def list(self) -> list[Measurand]:
        return super().list()

    @guard("view")
    def tabulate(self) -> pd.DataFrame:
        return super().tabulate()



import pandas as pd
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.data.auth.decorators import guard
from ixmp4.db import filters
from ixmp4.db.filters import BaseFilter

from .. import abstract, types
from . import base


class EnumerateKwargs(
    abstract.annotations.HasRunIdFilter,
    abstract.annotations.HasTransactionIdFilter,
    total=False,
):
    _filter: BaseFilter


class Checkpoint(base.BaseModel):
    __tablename__ = "checkpoint"
    run__id: types.Integer = db.Column(
        db.Integer,
        db.ForeignKey("run.id"),
        nullable=False,
        index=True,
    )
    transaction__id: types.Integer = db.Column(
        db.Integer,
        db.ForeignKey("transaction.id"),
        nullable=True,
        index=True,
    )
    message: types.String


class CheckpointFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    sqla_model = Checkpoint

    run__id: filters.Integer
    transaction__id: filters.Integer
    message: filters.String

    def join(
        self, exc: db.sql.Select[tuple[Checkpoint]], session: db.Session | None = None
    ) -> db.sql.Select[tuple[Checkpoint]]:
        return exc


class CheckpointRepository(
    base.Creator[Checkpoint],
    base.Deleter[Checkpoint],
    base.Enumerator[Checkpoint],
    abstract.CheckpointRepository,
):
    model_class = Checkpoint
    filter_class = CheckpointFilter

    def add(self, run__id: int, transaction__id: int, message: str) -> Checkpoint:
        checkpoint = Checkpoint(
            run__id=run__id, transaction__id=transaction__id, message=message
        )
        self.session.add(checkpoint)
        return checkpoint

    @guard("edit")
    def create(self, run__id: int, message: str) -> Checkpoint:
        # only save a transaction reference on pgsql
        if self.backend.engine.dialect.name == "postgresql":
            latest_transaction_id = self.backend.transactions.latest().id
        else:
            latest_transaction_id = None
        # permission check
        run = self.backend.runs.get_by_id(run__id, _access_type="edit")

        return super().create(run.id, latest_transaction_id, message)

    @guard("edit")
    def delete(self, id: int) -> None:
        return super().delete(id)

    @guard("view")
    def get_by_id(self, id: int) -> Checkpoint:
        obj = self.session.get(self.model_class, id)

        if obj is None:
            raise Checkpoint.NotFound(id=id)

        return obj

    @guard("view")
    def list(self, /, **kwargs: Unpack[EnumerateKwargs]) -> list[Checkpoint]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(self, **kwargs: Unpack[EnumerateKwargs]) -> pd.DataFrame:
        return super().tabulate(**kwargs)



from collections.abc import Iterable
from datetime import datetime
from typing import TYPE_CHECKING, Any, Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from .. import base
from ..docs import DocsRepository
from .base import BackendBaseRepository

if TYPE_CHECKING:
    from . import EnumerateKwargs


class Variable(base.BaseModel, Protocol):
    """Variable data model."""

    name: str
    """Unique name of the Variable."""
    data: dict[str, list[float] | list[int] | list[str]]
    """Data stored in the Variable."""
    indexset_names: list[str] | None
    """List of the names of the IndexSets the Variable is bound to."""
    column_names: list[str] | None
    """List of the Variable's column names, if distinct from the IndexSet names."""

    run__id: int
    "Foreign unique integer id of a run."

    created_at: datetime
    "Creation date/time. TODO"
    created_by: str
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<Variable {self.id} name={self.name}>"


class VariableRepository(
    BackendBaseRepository[Variable],
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(
        self,
        run_id: int,
        name: str,
        constrained_to_indexsets: str | list[str] | None = None,
        column_names: list[str] | None = None,
    ) -> Variable:
        """Creates a Variable.

        Each column of the Variable needs to be constrained to an existing
        :class:`ixmp4.data.abstract.optimization.IndexSet`. These are specified by name
        and per default, these will be the column names. They can be overwritten by
        specifying `column_names`, which needs to specify a unique name for each column.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Variable is
            defined.
        name : str
            The unique name of the Variable.
        constrained_to_indexsets : str | list[str] | None = None
            One or a List of :class:`ixmp4.data.abstract.optimization.IndexSet` names
            that define the allowed contents of the Variable's columns. If None, no data
            can be added beyond `levels` and `marginals`!
        column_names: list[str] | None = None
            Optional list of names to use as column names. If given, overwrites the
            names inferred from `constrained_to_indexsets`.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Variable.NotUnique`:
            If the Variable with `name` already exists for the Run with `run_id`.
        :class:`ixmp4.core.exceptions.OptimizationItemUsageError`:
            If `column_names` are not unique or not enough names are given.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Variable`:
            The created Variable.
        """
        ...

    def delete(self, id: int) -> None:
        """Deletes a Variable.

        Parameters
        ----------
        id : int
            The unique integer id of the Variable.

        Raises
        ------
        :class:`ixmp4.data.abstract.Variable.NotFound`:
            If the Variable with `id` does not exist.
        :class:`ixmp4.data.abstract.Variable.DeletionPrevented`:
            If the Variable with `id` is used in the database, preventing it's deletion.
        """
        ...

    def get(self, run_id: int, name: str) -> Variable:
        """Retrieves a Variable.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this Variable is
            defined.
        name : str
            The name of the Variable.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Variable.NotFound`:
            If the Variable with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Variable`:
            The retrieved Variable.
        """
        ...

    def get_by_id(self, id: int) -> Variable:
        """Retrieves a Variable by its id.

        Parameters
        ----------
        id : int
            Unique integer id.

        Raises
        ------
        :class:`ixmp4.data.abstract.optimization.Variable.NotFound`.
            If the Variable with `id` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.optimization.Variable`:
            The retrieved Variable.
        """
        ...

    def list(self, **kwargs: Unpack["EnumerateKwargs"]) -> Iterable[Variable]:
        r"""Lists Variables by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter Variables as specified in
            `ixmp4.data.db.optimization.variable.filter.OptimizationVariableFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.optimization.Variable`]:
            List of Variables.
        """
        ...

    def tabulate(self, **kwargs: Unpack["EnumerateKwargs"]) -> pd.DataFrame:
        r"""Tabulate Variables by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter variables as specified in
            `ixmp4.data.db.optimization.variable.filter.OptimizationVariableFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
                - data
                - run__id
                - created_at
                - created_by
        """
        ...

    # TODO Question for Daniel: do variables need to allow adding data manually?
    # TODO Once present, state how to check which IndexSets are linked and which values
    # they permit
    # TODO Can we remove the r-marker?
    def add_data(self, id: int, data: dict[str, Any] | pd.DataFrame) -> None:
        r"""Adds data to a Variable.

        The data will be validated with the linked constrained
        :class:`ixmp4.data.abstract.optimization.IndexSet`s. For that, `data.keys()`
        must correspond to the names of the Variable's columns. Each column can only
        contain values that are in the linked `IndexSet.data`. Each row of entries
        must be unique. No values can be missing, `None`, or `NaN`. If `data.keys()`
        contains names already present in `Variable.data`, existing values will be
        overwritten.

        Parameters
        ----------
        id : int
            The id of the :class:`ixmp4.data.abstract.optimization.Variable`.
        data : dict[str, Any] | pandas.DataFrame
            The data to be added.

        Raises
        ------
        :class:`ixmp4.core.exceptions.OptimizationItemUsageError`:
            - If values are missing, `None`, or `NaN`
            - If values are not allowed based on constraints to `Indexset`s
            - If rows are not unique

        Returns
        -------
        None
        """
        ...

    def remove_data(
        self, id: int, data: dict[str, Any] | pd.DataFrame | None = None
    ) -> None:
        """Removes data from a Variable.

        Parameters
        ----------
        id : int
            The id of the :class:`ixmp4.data.abstract.optimization.Variable`.
        data : dict[str, Any] | pandas.DataFrame, optional
            The data to be removed. If specified, remove only specific data. This must
            specify all indexed columns. All other keys/columns are ignored. Otherwise,
            remove all data (the default).

        Returns
        -------
        None
        """
        ...



from ixmp4.data.db import filters as base
from ixmp4.data.db.run import Run
from ixmp4.db import Session, filters, sql, utils

from .model import IndexSet, IndexSetData


class RunFilter(base.RunFilter, metaclass=filters.FilterMeta):
    def join(
        self, exc: sql.Select[tuple[IndexSet]], session: Session | None = None
    ) -> sql.Select[tuple[IndexSet]]:
        if not utils.is_joined(exc, Run):
            exc = exc.join(Run, onclause=IndexSet.run__id == Run.id)
        return exc


class OptimizationIndexSetFilter(
    base.OptimizationIndexSetFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: sql.Select[tuple[IndexSet]], session: Session | None = None
    ) -> sql.Select[tuple[IndexSet]]:
        return exc


class OptimizationIndexSetDataFilter(
    base.OptimizationIndexSetDataFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: sql.Select[tuple[IndexSetData]], session: Session | None = None
    ) -> sql.Select[tuple[IndexSetData]]:
        return exc



from datetime import datetime
from typing import TYPE_CHECKING, ClassVar, cast

if TYPE_CHECKING:
    from ixmp4.data.backend.api import RestBackend

import pandas as pd
from pydantic import Field

# TODO Import this from typing when dropping support for Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from . import base
from .docs import Docs, DocsRepository


class RegionParent(base.BaseModel):
    id: int
    name: str
    hierarchy: str


class Region(base.BaseModel):
    NotFound: ClassVar = abstract.Region.NotFound
    NotUnique: ClassVar = abstract.Region.NotUnique
    DeletionPrevented: ClassVar = abstract.Region.DeletionPrevented

    id: int
    name: str
    hierarchy: str

    created_at: datetime | None = Field(None)
    created_by: str | None = Field(None)


class RegionDocsRepository(DocsRepository):
    model_class = Docs
    prefix = "docs/regions/"


class RegionRepository(
    base.Creator[Region],
    base.Deleter[Region],
    base.Retriever[Region],
    base.Enumerator[Region],
    abstract.RegionRepository,
):
    model_class = Region
    prefix = "regions/"

    def __init__(self, *args: Unpack[tuple["RestBackend"]]) -> None:
        super().__init__(*args)
        self.docs = RegionDocsRepository(self.backend)

    def create(self, name: str, hierarchy: str) -> Region:
        return super().create(name=name, hierarchy=hierarchy)

    def delete(self, id: int) -> None:
        super().delete(id)

    def get(self, name: str) -> Region:
        return super().get(name=name)

    def enumerate(
        self, **kwargs: Unpack[abstract.region.EnumerateKwargs]
    ) -> list[Region] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs: Unpack[abstract.region.EnumerateKwargs]) -> list[Region]:
        json = cast(abstract.annotations.IamcObjectFilterAlias, kwargs)
        return super()._list(json=json)

    def tabulate(
        self, **kwargs: Unpack[abstract.region.EnumerateKwargs]
    ) -> pd.DataFrame:
        json = cast(abstract.annotations.IamcObjectFilterAlias, kwargs)
        return super()._tabulate(json=json)



from functools import partial, reduce
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.db.base import BaseModel as RootBaseModel
from ixmp4.data.db.optimization.base import Reverter
from ixmp4.data.db.optimization.indexset.model import IndexSetVersion
from ixmp4.data.db.versions.model import DefaultVersionModel
from ixmp4.data.db.versions.repository import VersionRepository
from ixmp4.db import utils
from ixmp4.db.utils.revert import apply_transaction__id

from .utils import create_id_map_subquery

if TYPE_CHECKING:
    from ixmp4.data.db.optimization import (
        EquationRepository,
        IndexSet,
        ParameterRepository,
        TableRepository,
        VariableRepository,
    )


class BaseIndexSetAssociation(RootBaseModel):
    __abstract__ = True

    indexset__id: types.IndexSetId

    @db.declared_attr
    def indexset(cls) -> types.Mapped["IndexSet"]:
        return db.relationship("IndexSet")

    column_name: types.String = db.Column(db.String(255), nullable=True)


class BaseIndexSetAssociationVersion(DefaultVersionModel):
    __abstract__ = True

    indexset__id: db.MappedColumn[int] = db.Column(
        db.Integer, nullable=False, index=True
    )

    column_name: types.String = db.Column(db.String(255), nullable=True)


AssociationModelType = TypeVar("AssociationModelType", bound=BaseIndexSetAssociation)
AssociationVersionModelType = TypeVar(
    "AssociationVersionModelType", bound=BaseIndexSetAssociationVersion
)


class BaseIndexSetAssociationVersionRepository(
    VersionRepository[AssociationVersionModelType]
):
    model_class: type[AssociationVersionModelType]
    parent_version: type[DefaultVersionModel]

    _item_id_column: Literal[
        "equation__id", "parameter__id", "table__id", "variable__id"
    ]

    def select(
        self,
        transaction__id: int | None = None,
        valid: Literal["at_transaction", "after_transaction"] = "at_transaction",
        indexset__ids: list[int] | None = None,
        item__ids: list[int] | None = None,
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = (
            db.select(self.bundle)
            .join_from(
                self.model_class,
                IndexSetVersion,
                self.model_class.indexset__id == IndexSetVersion.id,
            )
            .join_from(
                self.model_class,
                self.parent_version,
                getattr(self.model_class, self._item_id_column)
                == self.parent_version.id,
            )
        )

        _apply_transaction__id = partial(
            apply_transaction__id, transaction__id=transaction__id, valid=valid
        )

        exc = reduce(
            _apply_transaction__id,
            {self.model_class, IndexSetVersion, self.parent_version},
            exc,
        )

        if indexset__ids is not None and item__ids is not None:
            exc = exc.where(
                db.and_(
                    self.model_class.indexset__id.in_(indexset__ids),
                    getattr(self.model_class, self._item_id_column).in_(item__ids),
                )
            )

        exc = utils.where_matches_kwargs(exc, model_class=self.model_class, **kwargs)
        return exc.distinct()


class BaseIndexSetAssociationRepository(
    Reverter[AssociationModelType],
    Generic[AssociationModelType, AssociationVersionModelType],
):
    model_class: type[AssociationModelType]
    versions: BaseIndexSetAssociationVersionRepository[AssociationVersionModelType]


RepoType = TypeVar(
    "RepoType",
    "EquationRepository",
    "ParameterRepository",
    "TableRepository",
    "VariableRepository",
)


class BaseIndexSetAssociationReverter(
    Reverter[AssociationModelType],
    Generic[AssociationModelType, AssociationVersionModelType],
):
    _associations: BaseIndexSetAssociationRepository[
        AssociationModelType, AssociationVersionModelType
    ]

    _item_id_column: Literal[
        "equation__id", "parameter__id", "table__id", "variable__id"
    ]

    def _revert_indexset_association(
        self, repo: RepoType, transaction__id: int, run__id: int
    ) -> None:
        # Create subquery to map IDs from before rollback to new ones
        item_map_subquery = create_id_map_subquery(
            transaction__id=transaction__id, run__id=run__id, repo=repo
        )
        indexset_map_subquery = create_id_map_subquery(
            transaction__id=transaction__id,
            run__id=run__id,
            repo=self.backend.optimization.indexsets,
        )

        # Prepare columns with correctly updated IDs for SELECT
        columns = utils.collect_columns_to_select(
            columns=utils.get_columns(self._associations.versions.model_class),
            exclude={self._item_id_column, "indexset__id"},
        )

        # Construct SELECT to find correct versions with updated IDs
        select_correct_versions = (
            db.select(
                item_map_subquery.c.new_id.label(self._item_id_column),
                indexset_map_subquery.c.new_id.label("indexset__id"),
                *columns.values(),
            )
            .select_from(self._associations.versions.model_class)
            .join(
                item_map_subquery,
                getattr(self._associations.versions.model_class, self._item_id_column)
                == item_map_subquery.c.old_id,
            )
            .join(
                indexset_map_subquery,
                self._associations.versions.model_class.indexset__id
                == indexset_map_subquery.c.old_id,
            )
        )
        select_correct_versions = apply_transaction__id(
            exc=select_correct_versions,
            model_class=self._associations.versions.model_class,
            transaction__id=transaction__id,
        ).order_by(self._associations.versions.model_class.id.asc())

        correct_versions = self.tabulate_query(select_correct_versions)

        self._associations.revert(
            transaction__id=transaction__id, correct_versions=correct_versions
        )



# TODO Use `type` instead of TypeAlias when dropping Python 3.11
from datetime import datetime
from typing import Any, ClassVar, TypeAlias, cast

import pandas as pd
from pydantic import Field

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4.data import abstract

from . import base
from .model import Model
from .scenario import Scenario


class Run(base.BaseModel):
    NotFound: ClassVar = abstract.Run.NotFound
    NotUnique: ClassVar = abstract.Run.NotUnique
    DeletionPrevented: ClassVar = abstract.Run.DeletionPrevented

    NoDefaultVersion: ClassVar = abstract.Run.NoDefaultVersion

    id: int

    model: Model
    id_of_model: int = Field(..., alias="model__id")

    scenario: Scenario
    scenario__id: int

    version: int
    is_default: bool

    created_at: datetime | None
    created_by: str | None

    updated_at: datetime | None
    updated_by: str | None

    lock_transaction: int | None = None


JsonType: TypeAlias = dict[
    str,
    bool
    | abstract.annotations.IntFilterAlias
    | dict[
        str,
        abstract.annotations.DefaultFilterAlias
        | dict[str, abstract.annotations.DefaultFilterAlias],
    ]
    | None,
]


class RunRepository(
    base.Creator[Run],
    base.Deleter[Run],
    base.Retriever[Run],
    base.Enumerator[Run],
    abstract.RunRepository,
):
    model_class = Run
    prefix = "runs/"

    def create(self, model_name: str, scenario_name: str) -> Run:
        return super().create(model_name=model_name, scenario_name=scenario_name)

    def delete(self, id: int) -> None:
        super().delete(id)

    def get(self, model_name: str, scenario_name: str, version: int) -> Run:
        return super().get(
            model={"name": model_name},
            scenario={"name": scenario_name},
            version=version,
            default_only=False,
            is_default=None,
        )

    def get_by_id(self, id: int) -> Run:
        res = self._get_by_id(id)
        return Run(**res)

    def enumerate(
        self, **kwargs: Unpack[abstract.run.EnumerateKwargs]
    ) -> list[Run] | pd.DataFrame:
        return super().enumerate(**kwargs)

    def list(self, **kwargs: Unpack[abstract.run.EnumerateKwargs]) -> list[Run]:
        json = cast(JsonType, kwargs)
        return super()._list(json=json)

    def tabulate(self, **kwargs: Unpack[abstract.run.EnumerateKwargs]) -> pd.DataFrame:
        json = cast(JsonType, kwargs)
        return super()._tabulate(json=json)

    def get_default_version(self, model_name: str, scenario_name: str) -> Run:
        try:
            return super().get(
                model={"name": model_name},
                scenario={"name": scenario_name},
                is_default=True,
            )
        except Run.NotFound:
            raise Run.NoDefaultVersion

    def set_as_default_version(self, id: int) -> None:
        self._request(
            "POST",
            self.prefix + "/".join([str(id), "set-as-default-version/"]),
        )

    def unset_as_default_version(self, id: int) -> None:
        self._request(
            "POST",
            self.prefix + "/".join([str(id), "unset-as-default-version/"]),
        )

    def revert(
        self, id: int, transaction__id: int, revert_platform: bool = False
    ) -> None:
        self._request(
            "POST",
            self.prefix + "/".join([str(id), "revert/"]),
            json={
                "transaction__id": transaction__id,
                "revert_platform": revert_platform,
            },
        )

    def lock(self, id: int) -> Run:
        run_dict = self._request(
            "POST",
            self.prefix + "/".join([str(id), "lock/"]),
        )
        return Run(**cast(dict[str, Any], run_dict))

    def unlock(self, id: int) -> Run:
        run_dict = self._request(
            "POST",
            self.prefix + "/".join([str(id), "unlock/"]),
        )
        return Run(**cast(dict[str, Any], run_dict))

    def clone(
        self,
        run_id: int,
        model_name: str | None = None,
        scenario_name: str | None = None,
        keep_solution: bool = True,
    ) -> Run:
        # Can expect this endpoint, so cast should always be fine
        res = cast(
            dict[str, Any],
            self._request(
                "POST",
                self.prefix + "clone/",
                json={
                    "run_id": run_id,
                    "model_name": model_name,
                    "scenario_name": scenario_name,
                    "keep_solution": keep_solution,
                },
            ),
        )
        return Run(**res)



from typing import Any

from ixmp4.data.db.docs import BaseDocsRepository, docs_model

from .model import Parameter

ParameterDocs = docs_model(Parameter)


class ParameterDocsRepository(BaseDocsRepository[Any]):
    model_class = ParameterDocs
    dimension_model_class = Parameter



from typing import Any

from ixmp4 import db
from ixmp4.data.db import filters as base
from ixmp4.data.db.iamc.timeseries import TimeSeries
from ixmp4.db import Session, filters, typing_column, utils

from . import Region


class BaseIamcFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    def join_datapoints(
        self, exc: db.sql.Select[tuple[Region]], session: Session | None = None
    ) -> db.sql.Select[tuple[Region]]:
        if not utils.is_joined(exc, TimeSeries):
            exc = exc.join(TimeSeries, onclause=TimeSeries.region__id == Region.id)
        return exc


class SimpleIamcRegionFilter(
    base.RegionFilter, BaseIamcFilter, metaclass=filters.FilterMeta
):
    def join(
        self, exc: db.sql.Select[tuple[Region]], session: Session | None = None
    ) -> db.sql.Select[tuple[Region]]:
        return super().join_datapoints(exc, session)


class IamcRegionFilter(
    base.RegionFilter,
    BaseIamcFilter,
    metaclass=filters.FilterMeta,
):
    variable: base.VariableFilter | None = filters.Field(None)
    unit: base.UnitFilter | None = filters.Field(None)
    run: base.RunFilter = filters.Field(
        default=base.RunFilter(id=None, version=None, is_default=True)
    )

    _remote_filters = {"run", "variable", "unit"}
    _remote_path = [
        {
            "target_model": TimeSeries,
            "fk_attr": "region__id",
            "source_model": Region,
            "pk_attr": "id",
        },
    ]

    def join(
        self, exc: db.sql.Select[tuple[Region]], session: Session | None = None
    ) -> db.sql.Select[tuple[Region]]:
        if self._should_use_subquery_optimization():
            return exc

        return super().join_datapoints(exc, session)


class RegionFilter(base.RegionFilter, BaseIamcFilter, metaclass=filters.FilterMeta):
    iamc: IamcRegionFilter | filters.Boolean | None = filters.Field(None)

    def filter_iamc(
        self,
        exc: db.sql.Select[tuple[Region]],
        c: typing_column[Any],  # Any since it is unused
        v: bool | None,
        session: Session | None = None,
    ) -> db.sql.Select[tuple[Region]]:
        if v is None:
            return exc

        if v is True:
            return self.join_datapoints(exc, session)
        else:
            ids = self.join_datapoints(db.select(Region.id), session)
            exc = exc.where(~Region.id.in_(ids))
            return exc

    def join(
        self, exc: db.sql.Select[tuple[Region]], session: Session | None = None
    ) -> db.sql.Select[tuple[Region]]:
        return exc



# type: ignore
"""Make Scalar.unit non-optional

Revision ID: d66a8276ba0a
Revises: 914991d09f59
Create Date: 2024-11-29 14:14:42.857695

"""

import sqlalchemy as sa
from alembic import op

# Revision identifiers, used by Alembic.
revision = "d66a8276ba0a"
down_revision = "914991d09f59"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.alter_column("unit__id", existing_type=sa.INTEGER(), nullable=False)

    with op.batch_alter_table("runmetaentry", schema=None) as batch_op:
        batch_op.alter_column(
            "type",
            new_column_name="dtype",
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("runmetaentry", schema=None) as batch_op:
        batch_op.alter_column(
            "dtype",
            new_column_name="type",
        )

    with op.batch_alter_table("optimization_scalar", schema=None) as batch_op:
        batch_op.alter_column("unit__id", existing_type=sa.INTEGER(), nullable=True)

    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Any, cast

import numpy as np
import pandas as pd
import pandera as pa
from pandera.engines import pandas_engine
from pandera.pandas import DataFrameModel
from pandera.typing import DataFrame, Series

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from ixmp4 import db
from ixmp4.core.decorators import check_types
from ixmp4.core.exceptions import InconsistentIamcType, ProgrammingError
from ixmp4.data import abstract
from ixmp4.data.auth.decorators import guard
from ixmp4.data.db import versions
from ixmp4.data.db.model import Model
from ixmp4.data.db.region import Region, RegionVersion
from ixmp4.data.db.run import Run, RunRepository
from ixmp4.data.db.scenario import Scenario
from ixmp4.data.db.unit import Unit, UnitVersion
from ixmp4.db.filters import BaseFilter
from ixmp4.db.utils.revert import where_valid_at_transaction

from .. import base
from ..measurand import Measurand, MeasurandVersion
from ..timeseries import TimeSeries, TimeSeriesRepository, TimeSeriesVersion
from ..variable import Variable, VariableVersion
from . import get_datapoint_model
from .filter import DataPointFilter
from .model import DataPoint, UniversalDataPointVersion

if TYPE_CHECKING:
    from ixmp4.data.backend.db import SqlAlchemyBackend


class RemoveDataPointFrameSchema(DataFrameModel):
    type: Series[pa.String] | None = pa.Field(isin=[t for t in DataPoint.Type])
    step_year: Series[pa.Int] | None = pa.Field(coerce=True, nullable=True)
    step_datetime: Series[pandas_engine.DateTime] | None = pa.Field(
        coerce=True, nullable=True
    )
    step_category: Series[pa.String] | None = pa.Field(nullable=True)

    time_series__id: Series[pa.Int] = pa.Field(coerce=True)

    @pa.dataframe_check
    def validate_type(cls, df: pd.DataFrame) -> bool:
        types = df.type.unique()

        # mixed types are currently not supported
        if len(types) != 1:
            raise InconsistentIamcType

        if types[0] == DataPoint.Type.ANNUAL:
            cols = dict(step_year="valid", step_category="empty", step_datetime="empty")
        elif types[0] == DataPoint.Type.CATEGORICAL:
            cols = dict(step_year="valid", step_category="valid", step_datetime="empty")
        elif types[0] == DataPoint.Type.DATETIME:
            cols = dict(step_year="empty", step_category="empty", step_datetime="valid")
        else:
            raise ProgrammingError

        for col, content in cols.items():
            if infer_content(df, col) is not content:
                raise InconsistentIamcType
        return True


class AddDataPointFrameSchema(RemoveDataPointFrameSchema):
    value: Series[pa.Float] = pa.Field(coerce=True)


class UpdateDataPointFrameSchema(AddDataPointFrameSchema):
    id: Series[pa.Int] = pa.Field(coerce=True)


def infer_content(df: pd.DataFrame, col: str) -> str:
    if col not in df.columns or all(pd.isna(df[col])):
        return "empty"
    elif not any(pd.isna(df[col])):
        return "valid"
    else:
        raise InconsistentIamcType


class EnumerateKwargs(abstract.iamc.datapoint.EnumerateKwargs, total=False):
    join_parameters: bool | None
    join_runs: bool
    join_run_id: bool
    _filter: BaseFilter


class DatapointVersionRepository(versions.VersionRepository[UniversalDataPointVersion]):
    model_class = UniversalDataPointVersion

    def select(
        self,
        transaction__id: int | None = None,
        run__id: int | None = None,
        **kwargs: Any,
    ) -> db.sql.Select[Any]:
        exc = db.select(
            self.bundle,
            RegionVersion.name.label("region"),
            UnitVersion.name.label("unit"),
            VariableVersion.name.label("variable"),
        ).select_from(self.model_class)

        exc = (
            exc.join(
                TimeSeriesVersion,
                onclause=self.model_class.time_series__id == TimeSeriesVersion.id,
            )
            .join(
                RegionVersion,
                onclause=TimeSeriesVersion.region__id == RegionVersion.id,
            )
            .join(
                MeasurandVersion,
                onclause=TimeSeriesVersion.measurand__id == MeasurandVersion.id,
            )
            .join(
                UnitVersion,
                onclause=MeasurandVersion.unit__id == UnitVersion.id,
            )
            .join(
                VariableVersion,
                onclause=MeasurandVersion.variable__id == VariableVersion.id,
            )
        )

        if transaction__id is not None:
            for vclass in {
                self.model_class,
                RegionVersion,
                MeasurandVersion,
                UnitVersion,
                VariableVersion,
            }:
                exc = where_valid_at_transaction(exc, transaction__id, vclass)

        if run__id is not None:
            exc = exc.where(TimeSeriesVersion.run__id == run__id)

        exc = db.utils.where_matches_kwargs(exc, model_class=self.model_class, **kwargs)
        return exc.distinct()


class DataPointRepository(
    base.Enumerator[DataPoint],
    base.BulkUpserter[DataPoint],
    base.BulkDeleter[DataPoint],
    abstract.DataPointRepository,
):
    model_class = DataPoint
    timeseries: TimeSeriesRepository
    runs: RunRepository
    versions: DatapointVersionRepository

    def __init__(self, *args: "SqlAlchemyBackend") -> None:
        backend, *_ = args
        # A different table was used for ORACLE databases (deprecated since ixmp4 0.3.0)
        self.model_class = get_datapoint_model(backend.session)

        self.timeseries = TimeSeriesRepository(*args)
        self.runs = RunRepository(*args)
        self.versions = DatapointVersionRepository(*args)

        self.filter_class = DataPointFilter
        super().__init__(*args)

    def join_auth(
        self, exc: db.sql.Select[tuple[DataPoint]]
    ) -> db.sql.Select[tuple[DataPoint]]:
        if not db.utils.is_joined(exc, TimeSeries):
            exc = exc.join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
        if not db.utils.is_joined(exc, Run):
            exc = exc.join(Run, TimeSeries.run)
        if not db.utils.is_joined(exc, Model):
            exc = exc.join(Model, Run.model)

        return exc

    def select_joined_parameters(
        self, join_runs: bool = False, join_run_id: bool = False
    ) -> db.sql.Select[tuple[DataPoint]]:
        # NOTE Not quite sure about this bundle, seems to possibly take all types of
        # all model classes?
        bundle: list[db.Label[str] | db.Label[int] | db.Bundle[Any]] = []
        if join_runs:
            col_bundle: list[Any] = [
                Model.name.label("model"),
                Scenario.name.label("scenario"),
                Run.version.label("version"),
            ]
            if join_run_id:
                col_bundle.insert(0, Run.id.label("run__id"))
            bundle.extend(col_bundle)
        elif join_run_id:
            bundle.append(Run.id.label("run__id"))

        bundle.extend(
            [
                Region.name.label("region"),
                Unit.name.label("unit"),
                Variable.name.label("variable"),
                self.bundle,
            ]
        )

        _exc = (
            db.select(*bundle)
            .join(
                TimeSeries, onclause=self.model_class.time_series__id == TimeSeries.id
            )
            .join(Region, onclause=TimeSeries.region__id == Region.id)
            .join(Measurand, onclause=TimeSeries.measurand__id == Measurand.id)
            .join(Unit, onclause=Measurand.unit__id == Unit.id)
            .join(Variable, onclause=Measurand.variable__id == Variable.id)
        )

        if join_runs or join_run_id:
            _exc = _exc.join(Run, onclause=TimeSeries.run__id == Run.id)
            if join_runs:
                _exc = _exc.join(Model, onclause=Model.id == Run.model__id).join(
                    Scenario, onclause=Scenario.id == Run.scenario__id
                )
        return _exc

    def select(
        self,
        *,
        join_parameters: bool | None = False,
        join_runs: bool = False,
        join_run_id: bool = False,
        _filter: DataPointFilter | None = None,
        _exc: db.sql.Select[tuple[DataPoint]] | None = None,
        **kwargs: Any,
    ) -> db.sql.Select[tuple[DataPoint]]:
        if _exc is not None:
            exc = _exc
        elif join_parameters:
            exc = self.select_joined_parameters(join_runs, join_run_id)
        else:
            exc = db.select(self.bundle)

        return super().select(_exc=exc, _filter=_filter, **kwargs)

    @guard("view")
    def list(self, **kwargs: Unpack[EnumerateKwargs]) -> list[DataPoint]:
        return super().list(**kwargs)

    @guard("view")
    def tabulate(
        self, *args: Any, _raw: bool | None = False, **kwargs: Unpack[EnumerateKwargs]
    ) -> pd.DataFrame:
        if _raw:
            return super().tabulate(*args, **kwargs)
        df = super().tabulate(*args, **kwargs)
        df["value"] = df["value"].astype(np.float64)
        df = df.sort_values(
            by=["time_series__id", "step_year", "step_category", "step_datetime"]
        )
        return df.dropna(axis="columns", how="all")

    def check_df_access(self, df: pd.DataFrame) -> None:
        if self.backend.auth_context is not None:
            ts_ids = cast(set[int], set(df["time_series__id"].unique().tolist()))
            self.timeseries.check_access(
                ts_ids,
                access_type="edit",
                run={"default_only": False, "is_default": None},
            )

    @check_types
    @guard("edit")
    def bulk_upsert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        super().bulk_upsert(df)

    @check_types
    @guard("edit")
    def bulk_insert(self, df: DataFrame[AddDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        super().bulk_insert(df)

    @check_types
    @guard("edit")
    def bulk_update(self, df: DataFrame[UpdateDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        super().bulk_update(df)

    @check_types
    @guard("edit")
    def bulk_delete(self, df: DataFrame[RemoveDataPointFrameSchema]) -> None:
        self.check_df_access(df)
        super().bulk_delete(df)
        self.delete_orphans()

    def delete_orphans(self) -> None:
        exc = db.select(TimeSeries).where(
            ~db.exists(
                db.select(self.model_class.id).where(
                    TimeSeries.id == self.model_class.time_series__id
                )
            )
        )
        orphan_ts = self.timeseries.tabulate_query(exc)
        self.timeseries.bulk_delete(orphan_ts)



"""
This module is responsible for everything database related.
Interfacing, migrating and more is all done here.

It uses `sqlalchemy <https://www.sqlalchemy.org/>`__ and
`alembic <https://alembic.sqlalchemy.org/en/latest/>`__ for database
management.

Migrations
----------

There is a development database at ``run/db.sqlite`` which is used for
generating migrations, nothing else. It can be manipulated with alembic
directly using these commands:

.. code:: bash

   # run all migrations until the current state is reached
   alembic upgrade head

   # run one migration forward
   alembic upgrade +1

   # run one migration backward
   alembic downgrade -1

   # autogenerate new migration (please choose a descriptive change message)
   alembic revision -m "<message>" --autogenerate

You will have to run all migrations before being able to create new ones
in the development database. Be sure to run ``ruff`` on newly created
migrations before committing them!

"""

from typing import Annotated

from sqlalchemy import (
    BinaryExpression,
    BindParameter,
    ColumnExpressionArgument,
    ForeignKey,
    Index,
    Label,
    Sequence,
    UniqueConstraint,
    and_,
    delete,
    exists,
    false,
    func,
    insert,
    null,
    or_,
    select,
    sql,
    update,
)
from sqlalchemy import Column as typing_column
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.exc import IntegrityError, MultipleResultsFound
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import (
    Bundle,
    MappedColumn,
    Relationship,
    Session,
    aliased,
    backref,
    declared_attr,
    mapped_column,
    relationship,
    validates,
)
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.types import *

from . import utils

Column = mapped_column
EquationIdType = Annotated[
    int,
    Column(
        Integer,
        ForeignKey("opt_equ.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
]
# NOTE By using two IndexSetIdTypes, one with ondelete="CASCADE" and one without, we
# enable deletion even when the IndexSet is used in IndexSetData, but prevent it when
# it's used anywhere else
IndexSetIdType = Annotated[
    int,
    Column(Integer, ForeignKey("opt_idx.id"), nullable=False, index=True),
]
IndexSet__IdType = Annotated[
    int,
    Column(
        Integer,
        ForeignKey("opt_idx.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
]
JsonType = JSON()
JsonType = JsonType.with_variant(JSONB(), "postgresql")

name_len = 255
NameType = Annotated[str, Column(String(name_len), nullable=False, unique=False)]
UniqueNameType = Annotated[str, Column(String(name_len), nullable=False, unique=True)]

OptimizationVariableIdType = Annotated[
    int,
    Column(
        Integer,
        ForeignKey("opt_var.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
]
ParameterIdType = Annotated[
    int,
    Column(
        Integer,
        ForeignKey("opt_par.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
]
RunIdType = Annotated[
    int,
    Column(Integer, ForeignKey("run.id"), nullable=False, index=True),
]
TableIdType = Annotated[
    int,
    Column(
        Integer,
        ForeignKey("opt_tab.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    ),
]
UsernameType = Annotated[str, Column(String(255), nullable=True)]



import operator
from collections.abc import Callable, Iterable
from types import GenericAlias, UnionType
from typing import (
    Any,
    ClassVar,
    Optional,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

from pydantic import BaseModel, ConfigDict, Field, ValidationError, model_validator
from pydantic.fields import FieldInfo

# TODO Import these from typing when dropping support for 3.10
from typing_extensions import Protocol, Self, TypedDict

from ixmp4 import db
from ixmp4.core.exceptions import BadFilterArguments, ProgrammingError

in_Type = TypeVar("in_Type")


class RemotePathStep(TypedDict):
    """Type definition for a single step in a remote filter join path."""

    target_model: type
    fk_attr: str
    source_model: type
    pk_attr: str


class RemoteFilterConfig(Protocol):
    """Protocol for classes that support remote filter optimization."""

    _remote_filters: set[str]
    _remote_path: list[RemotePathStep]


def in_(
    c: db.typing_column[in_Type], v: Iterable[in_Type] | db.BindParameter[in_Type]
) -> db.BinaryExpression[bool]:
    return c.in_(v)


def like(c: db.typing_column[str], v: str) -> db.BinaryExpression[bool]:
    return c.like(escape_wildcard(v), escape="\\")


def ilike(c: db.typing_column[str], v: str) -> db.BinaryExpression[bool]:
    return c.ilike(escape_wildcard(v), escape="\\")


def notlike(c: db.typing_column[str], v: str) -> db.BinaryExpression[bool]:
    return c.notlike(escape_wildcard(v), escape="\\")


def notilike(c: db.typing_column[str], v: str) -> db.BinaryExpression[bool]:
    return c.notilike(escape_wildcard(v), escape="\\")


def escape_wildcard(v: str) -> str:
    return v.replace("%", "\\%").replace("*", "%")


class Integer(int):
    """An explicit proxy type for `int`."""

    pass


class Float(float):
    """An explicit proxy type for `float`."""

    pass


class Id(int):
    """A no-op type for a reduced set of `Integer` lookups."""

    pass


class String(str):
    """An explicit proxy type for `str`."""

    pass


Boolean = bool

argument_seperator = "__"
filter_func_prefix = "filter_"
lookup_map: dict[type, dict[str, tuple[type, Callable[..., Any]]]] = {
    Id: {
        "__root__": (int, operator.eq),
        "in": (list[int], in_),
    },
    Float: {
        "__root__": (int, operator.eq),
        "in": (list[int], in_),
        "gt": (int, operator.gt),
        "lt": (int, operator.lt),
        "gte": (int, operator.ge),
        "lte": (int, operator.le),
    },
    Integer: {
        "__root__": (int, operator.eq),
        "in": (list[int], in_),
        "gt": (int, operator.gt),
        "lt": (int, operator.lt),
        "gte": (int, operator.ge),
        "lte": (int, operator.le),
    },
    Boolean: {
        "__root__": (bool, operator.eq),
    },
    String: {
        "__root__": (str, operator.eq),
        "in": (list[str], in_),
        "like": (str, like),
        "ilike": (str, ilike),
        "notlike": (str, notlike),
        "notilike": (str, notilike),
    },
}


def get_filter_func_name(n: str) -> str:
    return filter_func_prefix + n.strip()


def _ensure_str_list(any_list: list[Any]) -> list[str]:
    str_list: list[str] = []
    for item in any_list:
        if isinstance(item, str):
            str_list.append(item)
        else:
            raise ProgrammingError("Field argument `lookups` must be `list` of `str`.")

    return str_list


PydanticMeta: type = type(BaseModel)


# NOTE mypy seems to say PydanticMeta has type Any, don't see how we could change that
class FilterMeta(PydanticMeta):  # type: ignore[misc]
    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> type["BaseFilter"]:
        annots = namespace.get("__annotations__", {}).copy()
        for _name, annot in annots.items():
            if get_origin(annot) == ClassVar:
                continue
            cls.process_field(namespace, _name, annot)

        return cast(FilterMeta, super().__new__(cls, name, bases, namespace, **kwargs))

    @classmethod
    def build_lookups(
        cls, field_type: type
    ) -> dict[str, tuple[type, Callable[..., Any]]]:
        global lookup_map
        if field_type not in lookup_map.keys():
            if get_origin(field_type) in [Union, UnionType]:
                unified_types: tuple[type, ...] = get_args(field_type)
                ut_lookup_map = {
                    type_: cls.build_lookups(type_) for type_ in unified_types
                }
                all_lookup_names: set[str] = set()
                for tl in ut_lookup_map.values():
                    all_lookup_names |= set(tl.keys())

                lookups: dict[str, tuple[type, Callable[..., Any]]] = {}
                for lookup_name in all_lookup_names:
                    tuples = [
                        ut_lookup_map[type_][lookup_name]
                        for type_ in unified_types
                        if ut_lookup_map[type_].get(lookup_name, None) is not None
                    ]
                    types, _ = zip(*tuples)

                    def lookup_func(
                        c: db.typing_column[Any],
                        v: (
                            Integer
                            | Float
                            | Id
                            | String
                            | Boolean
                            | list[Integer | Float | Id | String | Boolean]
                        ),
                        tuples: list[tuple[type, Callable[..., Any]]] = tuples,
                    ) -> Any:
                        for t, lf in tuples:
                            # NOTE can't check isinstance(..., list[...]) directly, but
                            # all these cases call the same lf() anyway; skipping
                            if isinstance(t, GenericAlias):
                                return lf(c, v)
                            if isinstance(v, t):
                                return lf(c, v)
                        raise ProgrammingError

                    lookups[lookup_name] = (
                        # dynamic union types can't
                        # be done according to type checkers
                        Union[tuple(types)],  # type: ignore[assignment]
                        lookup_func,
                    )
                return lookups
            else:
                return {"__root__": (field_type, operator.eq)}
        else:
            return lookup_map[field_type]

    @classmethod
    def process_field(
        cls, namespace: dict[str, Any], field_name: str, field_type: type
    ) -> None:
        lookups = cls.build_lookups(field_type)
        field: FieldInfo | None = namespace.get(field_name, Field(default=None))

        if field is None:
            return

        namespace.setdefault(field_name, field)
        override_lookups: list[str] | None = None
        if isinstance(field.json_schema_extra, dict):
            jschema_lookups = field.json_schema_extra.get("lookups", None)
            # NOTE We can't `isinstance` parametrized generics. Nothing seems to utilize
            # `lookups`, though, so this should not worsen performance.
            if isinstance(jschema_lookups, list):
                override_lookups = _ensure_str_list(jschema_lookups)
        else:
            override_lookups = None
        if isinstance(override_lookups, list):
            lookups = {k: v for k, v in lookups.items() if k in override_lookups}
        else:  # override_lookups is None
            pass
        base_field_alias = str(field.alias) if field.alias else field_name

        cls.expand_lookups(
            field_name,
            lookups,
            namespace,
            base_field_alias,
        )

    @classmethod
    def expand_lookups(
        cls,
        name: str,
        lookups: dict[str, tuple[type, Callable[..., Any]]],
        namespace: dict[str, Any],
        base_field_alias: str | None = None,
    ) -> None:
        global argument_seperator
        for lookup_alias, (type_, func) in lookups.items():
            filter_name = (
                name
                if lookup_alias == "__root__"
                else name + argument_seperator + lookup_alias
            )

            namespace["__annotations__"][filter_name] = Optional[type_]
            func_name = get_filter_func_name(filter_name)

            FilterType = TypeVar("FilterType")

            def filter_func(
                self: Self,
                exc: db.sql.Select[tuple[FilterType, ...]],
                f: str,
                v: Integer | Float | Id | String | Boolean,
                func: Callable[
                    [str, Integer | Float | Id | String | Boolean],
                    db.ColumnExpressionArgument[bool],
                ] = func,
                session: db.Session | None = None,
            ) -> db.sql.Select[tuple[FilterType, ...]]:
                return exc.where(func(f, v))

            namespace.setdefault(func_name, filter_func)

            if (
                base_field_alias is not None
                and base_field_alias != name
                and lookup_alias != "__root__"
            ):
                field = namespace.get(
                    filter_name,
                    Field(
                        None, alias=base_field_alias + argument_seperator + lookup_alias
                    ),
                )
            else:
                field = namespace.get(filter_name, Field(None))
            # NOTE field should always be pydantic FieldInfo, convince type checker
            assert field is not None
            json_schema_extra = {"sqla_column": name}
            field.json_schema_extra = json_schema_extra
            field._attributes_set["json_schema_extra"] = json_schema_extra
            namespace[filter_name] = field


ExpandType = TypeVar("ExpandType", str, list[str])
FilterType = TypeVar("FilterType")


class BaseFilter(BaseModel, metaclass=FilterMeta):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        extra="forbid",
        validate_by_alias=True,
        validate_by_name=True,
    )
    sqla_model: ClassVar[type | None] = None

    @model_validator(mode="before")
    @classmethod
    def expand_simple_filters(
        cls, v: ExpandType | dict[str, ExpandType]
    ) -> dict[str, ExpandType]:
        return expand_simple_filter(v)

    def __init__(self, **data: Any) -> None:
        try:
            super().__init__(**data)

        except ValidationError as e:
            raise BadFilterArguments(model=e.title, errors=e.errors())

    def join(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        session: db.Session | None = None,
    ) -> db.sql.Select[tuple[FilterType]]:
        return exc

    def apply(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        if self._should_use_subquery_optimization():
            return self._apply_with_subquery_optimization(exc, model, session)

        return self._apply(exc, model, session)

    def _should_use_subquery_optimization(self) -> bool:
        """Check if subquery optimization should be used for this filter."""
        remote_filters = getattr(self, "_remote_filters", None)
        if not remote_filters:
            return False

        for name in remote_filters:
            field_info = self.__class__.model_fields.get(name)
            if field_info:
                value = getattr(self, name, field_info.get_default())
                if isinstance(value, BaseFilter) and self._is_filter_active(value):
                    return True
        return False

    def _is_filter_active(self, filter_value: "BaseFilter") -> bool:
        """Check if a filter has any active (non-None) values."""
        field_values = filter_value.model_dump(exclude_none=True)
        return bool(field_values)

    def _apply(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Standard apply logic."""
        for name, field_info in self.__class__.model_fields.items():
            value = getattr(self, name, field_info.get_default())
            if isinstance(value, BaseFilter):
                exc = self._apply_nested_filter(exc, value, session)
            elif value is not None:
                exc = self._apply_field_filter(exc, name, value, model, session)
        return exc.distinct()

    def _apply_without_distinct(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply filters without adding DISTINCT (for use in subqueries)."""
        for name, field_info in self.__class__.model_fields.items():
            value = getattr(self, name, field_info.get_default())
            if isinstance(value, BaseFilter):
                exc = self._apply_nested_filter(exc, value, session)
            elif value is not None:
                exc = self._apply_field_filter(exc, name, value, model, session)
        return exc

    def _apply_with_subquery_optimization(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply filters using subquery optimization for remote filters."""
        remote_filters: set[str] = getattr(self, "_remote_filters", set())
        remote_path: list[RemotePathStep] | None = getattr(self, "_remote_path", None)
        if not remote_path:
            raise ProgrammingError(
                f"Filter {self.__class__.__name__} defines _remote_filters "
                "but missing _remote_path"
            )
        if model is None:
            return self._apply(exc, model, session)
        active_remote_filters = self._collect_active_remote_filters(remote_filters)
        if active_remote_filters:
            exc = self._apply_subquery_filter(
                exc, model, session, active_remote_filters, remote_path
            )
        exc = self._apply_local_filters(exc, model, session, remote_filters)
        return exc.distinct()

    def _collect_active_remote_filters(
        self, remote_filters: set[str]
    ) -> dict[str, "BaseFilter"]:
        """Collect remote filters that are active."""
        active_remote_filters = {}
        for name in remote_filters:
            value = getattr(self, name, None)
            if isinstance(value, BaseFilter) and self._is_filter_active(value):
                active_remote_filters[name] = value
        return active_remote_filters

    def _get_model_id_column(self, model: object) -> Any:
        """Safely get the ID column from a model class."""
        if not hasattr(model, "id"):
            raise ProgrammingError(
                f"Model {model.__class__.__name__} missing required 'id' column"
            )
        id_column = getattr(model, "id")
        if not hasattr(id_column, "type"):
            raise ProgrammingError(
                f"Model {model.__class__.__name__}.id is not a SQLAlchemy column"
            )
        return id_column

    def _apply_subquery_filter(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
        active_remote_filters: dict[str, "BaseFilter"],
        remote_path: list[RemotePathStep],
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply remote filters using subquery."""
        current_id_column = self._get_model_id_column(model)
        subquery = db.sql.select(current_id_column)
        subquery = self._build_subquery_joins(subquery, remote_path)
        subquery = self._apply_filters_to_subquery(
            subquery, active_remote_filters, session
        )
        model_id_column = self._get_model_id_column(
            getattr(self, "sqla_model", None) or model
        )
        return exc.where(model_id_column.in_(subquery.scalar_subquery()))

    def _build_subquery_joins(
        self,
        subquery: db.sql.Select[tuple[FilterType]],
        remote_path: list[RemotePathStep],
    ) -> db.sql.Select[tuple[FilterType]]:
        """Build the join chain for the subquery."""
        for step in remote_path:
            target_model = step["target_model"]
            fk_attr = step["fk_attr"]
            source_model = step["source_model"]
            pk_attr = step["pk_attr"]

            fk_column = getattr(target_model, fk_attr)
            pk_column = getattr(source_model, pk_attr)
            join_condition = fk_column == pk_column
            subquery = subquery.join(target_model, join_condition)
        return subquery

    def _apply_filters_to_subquery(
        self,
        subquery: db.sql.Select[tuple[FilterType]],
        active_remote_filters: dict[str, "BaseFilter"],
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply remote filters to the subquery."""
        for filter_value in active_remote_filters.values():
            subquery = filter_value.join(subquery, session=session)
            # We don't want DISTINCT in the subquery
            # it is potentially very costly and not needed
            subquery = filter_value._apply_without_distinct(
                subquery, filter_value.sqla_model, session
            )
        return subquery

    def _apply_local_filters(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        model: object,
        session: db.Session,
        remote_filters: set[str],
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply local (non-remote) filters."""
        for name, field_info in self.__class__.model_fields.items():
            if name in remote_filters:
                continue
            value = getattr(self, name, field_info.get_default())
            if isinstance(value, BaseFilter):
                exc = self._apply_nested_filter(exc, value, session)
            elif value is not None:
                exc = self._apply_field_filter(exc, name, value, model, session)
        return exc

    def _apply_nested_filter(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        filter_value: "BaseFilter",
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply a nested filter."""
        submodel = getattr(filter_value, "sqla_model", None)
        model_getter = getattr(filter_value, "get_sqla_model", None)
        if submodel is None and callable(model_getter):
            submodel = model_getter(session)

        exc = filter_value.join(exc, session=session)
        return filter_value.apply(exc, submodel, session)

    def _apply_field_filter(
        self,
        exc: db.sql.Select[tuple[FilterType]],
        name: str,
        value: Any,
        model: object,
        session: db.Session,
    ) -> db.sql.Select[tuple[FilterType]]:
        """Apply a field-level filter."""
        func_name = get_filter_func_name(name)
        filter_func = getattr(self, func_name, None)
        if filter_func is None:
            raise ProgrammingError

        sqla_column = self._get_sqla_column(name)
        column = None if sqla_column is None else getattr(model, sqla_column, None)
        result = filter_func(exc, column, value, session=session)
        if not isinstance(result, db.sql.Select):
            raise ProgrammingError("Filter function must return a Select query")
        return result

    def _get_sqla_column(self, name: str) -> str | None:
        """Get the SQLAlchemy column name for a field."""
        field_info = self.__class__.model_fields.get(name)
        if field_info and isinstance(field_info.json_schema_extra, dict):
            jschema_col = field_info.json_schema_extra.get("sqla_column")
            if not isinstance(jschema_col, str):
                raise ProgrammingError(
                    "Field argument `sqla_column` must be of type `str`."
                )
            return jschema_col
        return None


def expand_simple_filter(
    value: ExpandType | dict[str, ExpandType],
) -> dict[str, ExpandType]:
    if isinstance(value, str):
        return dict(name__like=value) if "*" in value else dict(name=value)
    elif isinstance(value, list):
        if any(["*" in v for v in value]):
            raise NotImplementedError("Filter by list with wildcard is not implemented")
        return dict(name__in=value)

    return value



from collections.abc import Iterable

from ..annotations import HasIdFilter, HasNameFilter, HasRunIdFilter
from .base import BackendBaseRepository
from .equation import Equation, EquationRepository
from .indexset import IndexSet, IndexSetRepository
from .parameter import Parameter, ParameterRepository
from .scalar import Scalar, ScalarRepository
from .table import Table, TableRepository
from .variable import Variable, VariableRepository


class EnumerateKwargs(HasIdFilter, HasNameFilter, HasRunIdFilter, total=False): ...



# TODO Import this from typing when dropping Python 3.11
from typing_extensions import TypedDict

from ixmp4.data.backend import Backend

from .data import OptimizationData


class InitKwargs(TypedDict):
    _backend: Backend | None



from datetime import datetime
from typing import TYPE_CHECKING, List, Protocol

import pandas as pd

# TODO Import this from typing when dropping Python 3.11
from typing_extensions import Unpack

from .. import base
from ..docs import DocsRepository
from .base import BackendBaseRepository

if TYPE_CHECKING:
    from . import EnumerateKwargs


class IndexSet(base.BaseModel, Protocol):
    """IndexSet data model."""

    name: str
    """Unique name of the IndexSet."""
    run__id: int
    """The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
    defined. """

    data: list[float] | list[int] | list[str]
    """Unique list of str, int, or float."""

    created_at: datetime
    "Creation date/time. TODO"
    created_by: str
    "Creator. TODO"

    def __str__(self) -> str:
        return f"<IndexSet {self.id} name={self.name}>"


class IndexSetRepository(
    BackendBaseRepository[IndexSet],
    base.Creator,
    base.Deleter,
    base.Retriever,
    base.Enumerator,
    Protocol,
):
    docs: DocsRepository

    def create(self, run_id: int, name: str) -> IndexSet:
        """Creates an IndexSet.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
            defined.
        name : str
            The name of the IndexSet.

        Raises
        ------
        :class:`ixmp4.data.abstract.IndexSet.NotUnique`:
            If the IndexSet with `name` already exists.

        Returns
        -------
        :class:`ixmp4.data.abstract.IndexSet`:
            The created IndexSet.
        """
        ...

    def delete(self, id: int) -> None:
        """Deletes an IndexSet.

        Parameters
        ----------
        id : int
            The unique integer id of the IndexSet.

        Raises
        ------
        :class:`ixmp4.data.abstract.IndexSet.NotFound`:
            If the IndexSet with `id` does not exist.
        :class:`ixmp4.data.abstract.IndexSet.DeletionPrevented`:
            If the IndexSet with `id` is used in the database, preventing it's deletion.
        """
        ...

    def get(self, run_id: int, name: str) -> IndexSet:
        """Retrieves an IndexSet.

        Parameters
        ----------
        run_id : int
            The id of the :class:`ixmp4.data.abstract.Run` for which this IndexSet is
            defined.
        name : str
            The unique name of the IndexSet.

        Raises
        ------
        :class:`ixmp4.data.abstract.IndexSet.NotFound`:
            If the IndexSet with `name` does not exist.

        Returns
        -------
        :class:`ixmp4.data.abstract.IndexSet`:
            The retrieved IndexSet.
        """
        ...

    def list(self, **kwargs: Unpack["EnumerateKwargs"]) -> list[IndexSet]:
        r"""Lists IndexSets by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.optimization.indexset.filter.OptimizationIndexSetFilter`.

        Returns
        -------
        Iterable[:class:`ixmp4.data.abstract.IndexSet`]:
            List of IndexSets.
        """
        ...

    def tabulate(self, **kwargs: Unpack["EnumerateKwargs"]) -> pd.DataFrame:
        r"""Tabulate IndexSets by specified criteria.

        Parameters
        ----------
        \*\*kwargs: any
            Any filter parameters as specified in
            `ixmp4.data.db.optimization.indexset.filter.OptimizationIndexSetFilter`.

        Returns
        -------
        :class:`pandas.DataFrame`:
            A data frame with the columns:
                - id
                - name
                - data
                - run__id
                - created_at
                - created_by
        """
        ...

    def add_data(
        self, id: int, data: float | int | str | List[float] | List[int] | List[str]
    ) -> None:
        """Adds data to an existing IndexSet.

        Parameters
        ----------
        id : int
            The id of the target IndexSet.
        data : float | int | str | List[float] | List[int] | List[str]
            The data to be added to the IndexSet.

        Returns
        -------
        None:
            Due to compatibility with ixmp.
        """
        ...

    def remove_data(
        self,
        id: int,
        data: float | int | str | List[float] | List[int] | List[str],
        remove_dependent_data: bool = True,
    ) -> None:
        """Removes data from an existing IndexSet.

        Parameters
        ----------
        id : int
            The id of the target IndexSet.
        data : float | int | str | List[float] | List[int] | List[str]
            The data to be removed from the IndexSet.
        remove_dependent_data : bool, optional
            Whether to delete data from all linked items referencing `data`.
            Default: `True`.

        Returns
        -------
        None:
            Due to compatibility with ixmp.
        """
        ...



from fastapi import APIRouter, Body, Depends, Query

from ixmp4.data import api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.optimization.indexset.filter import OptimizationIndexSetFilter
from ixmp4.data.db.optimization.indexset.model import IndexSet

from .. import deps
from ..base import BaseModel, EnumerationOutput, Pagination
from ..decorators import autodoc

router: APIRouter = APIRouter(
    prefix="/indexsets",
    tags=["optimization", "indexsets"],
)


class IndexSetInput(BaseModel):
    run_id: int
    name: str


class DataInput(BaseModel):
    data: float | int | str | list[float] | list[int] | list[str]


@autodoc
@router.patch("/", response_model=EnumerationOutput[api.IndexSet])
def query(
    filter: OptimizationIndexSetFilter = Body(OptimizationIndexSetFilter()),
    table: bool = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[IndexSet]:
    return EnumerationOutput(
        results=backend.optimization.indexsets.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
        ),
        total=backend.optimization.indexsets.count(_filter=filter),
        pagination=pagination,
    )


@autodoc
@router.post("/", response_model=api.IndexSet)
def create(
    indexset: IndexSetInput,
    backend: Backend = Depends(deps.get_backend),
) -> IndexSet:
    return backend.optimization.indexsets.create(**indexset.model_dump())


@autodoc
@router.delete("/{id}/")
def delete(
    id: int,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.indexsets.delete(id=id)


@autodoc
@router.patch("/{id}/data/")
def add_data(
    id: int,
    data: DataInput,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.indexsets.add_data(id=id, **data.model_dump())


@autodoc
@router.delete("/{id}/data/")
def remove_data(
    id: int,
    data: DataInput,
    remove_dependent_data: bool = Query(True),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.optimization.indexsets.remove_data(
        id=id, remove_dependent_data=remove_dependent_data, **data.model_dump()
    )



from typing import TYPE_CHECKING

from ixmp4.data.backend import Backend

from ..base import BaseFacade
from .equation import EquationRepository
from .indexset import IndexSetRepository
from .parameter import ParameterRepository
from .scalar import ScalarRepository
from .table import TableRepository
from .variable import VariableRepository

if TYPE_CHECKING:
    from ixmp4.core.run import Run


class OptimizationData(BaseFacade):
    """An optimization data instance, which provides access to optimization data such as
    IndexSet, Table, Variable, etc."""

    equations: EquationRepository
    indexsets: IndexSetRepository
    parameters: ParameterRepository
    scalars: ScalarRepository
    tables: TableRepository
    variables: VariableRepository

    def __init__(self, run: "Run", **kwargs: Backend) -> None:
        super().__init__(**kwargs)
        self.equations = EquationRepository(_backend=self.backend, _run=run)
        self.indexsets = IndexSetRepository(_backend=self.backend, _run=run)
        self.parameters = ParameterRepository(_backend=self.backend, _run=run)
        self.scalars = ScalarRepository(_backend=self.backend, _run=run)
        self.tables = TableRepository(_backend=self.backend, _run=run)
        self.variables = VariableRepository(_backend=self.backend, _run=run)

    # TODO Improve performance by writing dedicated queries
    def remove_solution(self) -> None:
        for equation in self.equations.list():
            equation.remove_data()
        for variable in self.variables.list():
            variable.remove_data()

    def has_solution(self) -> bool:
        """Check whether this Run contains a solution."""
        for variable in self.variables.list():
            if variable.levels:
                return True
        for equation in self.equations.list():
            if equation.levels:
                return True

        # If neither variables nor equations contain data, it doesn't
        return False



from fastapi import APIRouter, Body, Depends, Path, Query

from ixmp4.core.exceptions import BadRequest
from ixmp4.data import abstract, api
from ixmp4.data.backend.db import SqlAlchemyBackend as Backend
from ixmp4.data.db.meta.filter import RunMetaEntryFilter
from ixmp4.data.db.meta.model import RunMetaEntry

from . import deps
from .base import BaseModel, EnumerationOutput, Pagination

router: APIRouter = APIRouter(
    prefix="/meta",
    tags=["meta"],
)


class RunMetaEntryInput(BaseModel):
    run__id: int
    key: str
    value: abstract.StrictMetaValue


@router.patch("/", response_model=EnumerationOutput[api.RunMetaEntry])
def query(
    filter: RunMetaEntryFilter = Body(None),
    join_run_index: bool | None = Query(False),
    table: bool | None = Query(False),
    pagination: Pagination = Depends(),
    backend: Backend = Depends(deps.get_backend),
) -> EnumerationOutput[RunMetaEntry]:
    if join_run_index and not table:
        raise BadRequest("`join_run_index` can only be used with `table=true`.")

    return EnumerationOutput(
        results=backend.meta.paginate(
            _filter=filter,
            limit=pagination.limit,
            offset=pagination.offset,
            table=bool(table),
            join_run_index=join_run_index,
        ),
        total=backend.meta.count(_filter=filter),
        pagination=pagination,
    )


@router.post("/", response_model=api.RunMetaEntry)
def create(
    runmeta: RunMetaEntryInput,
    backend: Backend = Depends(deps.get_backend),
) -> RunMetaEntry:
    return backend.meta.create(**runmeta.model_dump())


@router.delete("/{id}/")
def delete(
    id: int = Path(),
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.meta.delete(id)


@router.post("/bulk/")
def bulk_upsert(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    # A pandera.DataFrame is a subclass of pd.DataFrame, so this is fine. Mypy likely
    # complains because our decorators change the type hint in some incompatible way.
    # Might be about covariance again.
    backend.meta.bulk_upsert(df.to_pandas())  # type: ignore[arg-type]


@router.patch("/bulk/")
def bulk_delete(
    df: api.DataFrame,
    backend: Backend = Depends(deps.get_backend),
) -> None:
    backend.meta.bulk_delete(df.to_pandas())  # type: ignore[arg-type]



from ixmp4.data.db import filters as base
from ixmp4.db import Session, filters, sql

from .model import TimeSeries


class TimeSeriesFilter(base.TimeSeriesFilter, metaclass=filters.FilterMeta):
    run: base.RunFilter = filters.Field(default=base.RunFilter())
    region: base.RegionFilter | None = filters.Field(None)
    variable: base.VariableFilter | None = filters.Field(None)
    unit: base.UnitFilter | None = filters.Field(None)

    def join(
        self, exc: sql.Select[tuple[TimeSeries]], session: Session | None = None
    ) -> sql.Select[tuple[TimeSeries]]:
        return exc


class TimeSeriesVersionFilter(filters.BaseFilter, metaclass=filters.FilterMeta):
    id: filters.Id | None = filters.Field(None)
    run__id: filters.Id | None = filters.Field(None)
    transaction_id: filters.Id | None = filters.Field(None)
    end_transaction_id: filters.Id | None = filters.Field(None)
